#!/usr/bin/env python
"""A simple example demonstrating how to enable tab completion by assigning a completer function to do_* commands.
This also demonstrates capabilities of the following completer features included with cmd2:
- CompletionError exceptions
- delimiter_complete()
- flag_based_complete() (see note below)
- index_based_complete() (see note below).

flag_based_complete() and index_based_complete() are basic methods and should only be used if you are not
familiar with argparse. The recommended approach for tab completing positional tokens and flags is to use
argparse-based completion. For an example integrating tab completion with argparse, see argparse_completion.py
"""

import functools

import cmd2

# List of strings used with completion functions
food_item_strs = ['Pizza', 'Ham', 'Ham Sandwich', 'Potato']
sport_item_strs = ['Bat', 'Basket', 'Basketball', 'Football', 'Space Ball']

# This data is used to demonstrate delimiter_complete
file_strs = [
    '/home/user/file.db',
    '/home/user/file space.db',
    '/home/user/another.db',
    '/home/other user/maps.db',
    '/home/other user/tests.db',
]


class BasicCompletion(cmd2.Cmd):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def do_flag_based(self, statement: cmd2.Statement) -> None:
        """Tab completes arguments based on a preceding flag using flag_based_complete
        -f, --food [completes food items]
        -s, --sport [completes sports]
        -p, --path [completes local file system paths].
        """
        self.poutput(f"Args: {statement.args}")

    def complete_flag_based(self, text, line, begidx, endidx) -> list[str]:
        """Completion function for do_flag_based."""
        flag_dict = {
            # Tab complete food items after -f and --food flags in command line
            '-f': food_item_strs,
            '--food': food_item_strs,
            # Tab complete sport items after -s and --sport flags in command line
            '-s': sport_item_strs,
            '--sport': sport_item_strs,
            # Tab complete using path_complete function after -p and --path flags in command line
            '-p': self.path_complete,
            '--path': self.path_complete,
        }

        return self.flag_based_complete(text, line, begidx, endidx, flag_dict=flag_dict)

    def do_index_based(self, statement: cmd2.Statement) -> None:
        """Tab completes first 3 arguments using index_based_complete."""
        self.poutput(f"Args: {statement.args}")

    def complete_index_based(self, text, line, begidx, endidx) -> list[str]:
        """Completion function for do_index_based."""
        index_dict = {
            1: food_item_strs,  # Tab complete food items at index 1 in command line
            2: sport_item_strs,  # Tab complete sport items at index 2 in command line
            3: self.path_complete,  # Tab complete using path_complete function at index 3 in command line
        }

        return self.index_based_complete(text, line, begidx, endidx, index_dict=index_dict)

    def do_delimiter_complete(self, statement: cmd2.Statement) -> None:
        """Tab completes files from a list using delimiter_complete."""
        self.poutput(f"Args: {statement.args}")

    # Use a partialmethod to set arguments to delimiter_complete
    complete_delimiter_complete = functools.partialmethod(cmd2.Cmd.delimiter_complete, match_against=file_strs, delimiter='/')

    def do_raise_error(self, statement: cmd2.Statement) -> None:
        """Demonstrates effect of raising CompletionError."""
        self.poutput(f"Args: {statement.args}")

    def complete_raise_error(self, _text, _line, _begidx, _endidx) -> list[str]:
        """CompletionErrors can be raised if an error occurs while tab completing.

        Example use cases
            - Reading a database to retrieve a tab completion data set failed
            - A previous command line argument that determines the data set being completed is invalid
        """
        raise cmd2.CompletionError("This is how a CompletionError behaves")


if __name__ == '__main__':
    import sys

    app = BasicCompletion()
    sys.exit(app.cmdloop())



#!/usr/bin/env python
"""A sample application for cmd2. Demonstrating colors available in the cmd2.colors.Color enum.

Execute the taste_the_rainbow command to see the colors available.
"""

import argparse

from rich.style import Style
from rich.text import Text

import cmd2
from cmd2 import Color


class CmdLineApp(cmd2.Cmd):
    """Example cmd2 application demonstrating colorized output."""

    def __init__(self) -> None:
        # Set include_ipy to True to enable the "ipy" command which runs an interactive IPython shell
        super().__init__(include_ipy=True)
        self.intro = 'Run the taste_the_rainbow command to see all of the colors available to you in cmd2.'

    rainbow_parser = cmd2.Cmd2ArgumentParser()
    rainbow_parser.add_argument('-b', '--background', action='store_true', help='show background colors as well')
    rainbow_parser.add_argument('-p', '--paged', action='store_true', help='display output using a pager')

    @cmd2.with_argparser(rainbow_parser)
    def do_taste_the_rainbow(self, args: argparse.Namespace) -> None:
        """Show all of the colors available within cmd2's Color StrEnum class."""

        def create_style(color: Color) -> Style:
            """Create a foreground or background color Style."""
            if args.background:
                return Style(bgcolor=color)
            return Style(color=color)

        styled_names = [Text(color.name, style=create_style(color)) for color in Color]
        output = Text("\n").join(styled_names)

        if args.paged:
            self.ppaged(output)
        else:
            self.poutput(output)


if __name__ == '__main__':
    import sys

    c = CmdLineApp()
    sys.exit(c.cmdloop())



#!/usr/bin/env python3
"""A simple example demonstrating how to pretty print JSON data in a cmd2 app using rich."""

from rich.json import JSON

import cmd2

EXAMPLE_DATA = {
    "name": "John Doe",
    "age": 30,
    "address": {"street": "123 Main St", "city": "Anytown", "state": "CA"},
    "hobbies": ["reading", "hiking", "coding"],
}


class Cmd2App(cmd2.Cmd):
    def __init__(self) -> None:
        super().__init__()
        self.data = EXAMPLE_DATA

    def do_normal(self, _) -> None:
        """Display the data using the normal poutput method."""
        self.poutput(self.data)

    def do_pretty(self, _) -> None:
        """Display the JSON data in a pretty way using rich."""

        json_renderable = JSON.from_data(
            self.data,
            indent=2,
            highlight=True,
            skip_keys=False,
            ensure_ascii=False,
            check_circular=True,
            allow_nan=True,
            default=None,
            sort_keys=False,
        )
        self.poutput(json_renderable)


if __name__ == '__main__':
    app = Cmd2App()
    app.cmdloop()



#!/usr/bin/env python
"""A simple example demonstrating the various ways to call cmd2.Cmd.read_input() for input history and tab completion."""

import contextlib

import cmd2

EXAMPLE_COMMANDS = "Example Commands"


class ReadInputApp(cmd2.Cmd):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.prompt = "\n" + self.prompt
        self.custom_history = ['history 1', 'history 2']

    @cmd2.with_category(EXAMPLE_COMMANDS)
    def do_basic(self, _) -> None:
        """Call read_input with no history or tab completion."""
        self.poutput("Tab completion and up-arrow history is off")
        with contextlib.suppress(EOFError):
            self.read_input("> ")

    @cmd2.with_category(EXAMPLE_COMMANDS)
    def do_basic_with_history(self, _) -> None:
        """Call read_input with custom history and no tab completion."""
        self.poutput("Tab completion is off but using custom history")
        try:
            input_str = self.read_input("> ", history=self.custom_history)
        except EOFError:
            pass
        else:
            self.custom_history.append(input_str)

    @cmd2.with_category(EXAMPLE_COMMANDS)
    def do_commands(self, _) -> None:
        """Call read_input the same way cmd2 prompt does to read commands."""
        self.poutput("Tab completing and up-arrow history configured for commands")
        with contextlib.suppress(EOFError):
            self.read_input("> ", completion_mode=cmd2.CompletionMode.COMMANDS)

    @cmd2.with_category(EXAMPLE_COMMANDS)
    def do_custom_choices(self, _) -> None:
        """Call read_input to use custom history and choices."""
        self.poutput("Tab completing with static choices list and using custom history")
        try:
            input_str = self.read_input(
                "> ",
                history=self.custom_history,
                completion_mode=cmd2.CompletionMode.CUSTOM,
                choices=['choice_1', 'choice_2', 'choice_3'],
            )
        except EOFError:
            pass
        else:
            self.custom_history.append(input_str)

    def choices_provider(self) -> list[str]:
        """Example choices provider function."""
        return ["from_provider_1", "from_provider_2", "from_provider_3"]

    @cmd2.with_category(EXAMPLE_COMMANDS)
    def do_custom_choices_provider(self, _) -> None:
        """Call read_input to use custom history and choices provider function."""
        self.poutput("Tab completing with choices from provider function and using custom history")
        try:
            input_str = self.read_input(
                "> ",
                history=self.custom_history,
                completion_mode=cmd2.CompletionMode.CUSTOM,
                choices_provider=ReadInputApp.choices_provider,
            )
        except EOFError:
            pass
        else:
            self.custom_history.append(input_str)

    @cmd2.with_category(EXAMPLE_COMMANDS)
    def do_custom_completer(self, _) -> None:
        """Call read_input to use custom history and completer function."""
        self.poutput("Tab completing paths and using custom history")
        try:
            input_str = self.read_input(
                "> ", history=self.custom_history, completion_mode=cmd2.CompletionMode.CUSTOM, completer=cmd2.Cmd.path_complete
            )
            self.custom_history.append(input_str)
        except EOFError:
            pass

    @cmd2.with_category(EXAMPLE_COMMANDS)
    def do_custom_parser(self, _) -> None:
        """Call read_input to use a custom history and an argument parser."""
        parser = cmd2.Cmd2ArgumentParser(prog='', description="An example parser")
        parser.add_argument('-o', '--option', help="an optional arg")
        parser.add_argument('arg_1', help="a choice for this arg", metavar='arg_1', choices=['my_choice', 'your_choice'])
        parser.add_argument('arg_2', help="path of something", completer=cmd2.Cmd.path_complete)

        self.poutput("Tab completing with argument parser and using custom history")
        self.poutput(parser.format_usage())

        try:
            input_str = self.read_input(
                "> ", history=self.custom_history, completion_mode=cmd2.CompletionMode.CUSTOM, parser=parser
            )
        except EOFError:
            pass
        else:
            self.custom_history.append(input_str)


if __name__ == '__main__':
    import sys

    app = ReadInputApp()
    sys.exit(app.cmdloop())



"""Unit/functional testing for argparse customizations in cmd2"""

import argparse

import pytest

import cmd2
from cmd2 import (
    Cmd2ArgumentParser,
    constants,
)
from cmd2.argparse_custom import generate_range_error

from .conftest import run_cmd


class ApCustomTestApp(cmd2.Cmd):
    """Test app for cmd2's argparse customization"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    range_parser = Cmd2ArgumentParser()
    range_parser.add_argument('--arg0', nargs=1)
    range_parser.add_argument('--arg1', nargs=2)
    range_parser.add_argument('--arg2', nargs=(3,))
    range_parser.add_argument('--arg3', nargs=(2, 3))

    @cmd2.with_argparser(range_parser)
    def do_range(self, _) -> None:
        pass


@pytest.fixture
def cust_app():
    return ApCustomTestApp()


def fake_func() -> None:
    pass


@pytest.mark.parametrize(
    ('kwargs', 'is_valid'),
    [
        ({'choices_provider': fake_func}, True),
        ({'completer': fake_func}, True),
        ({'choices_provider': fake_func, 'completer': fake_func}, False),
    ],
)
def test_apcustom_choices_callable_count(kwargs, is_valid) -> None:
    parser = Cmd2ArgumentParser()
    if is_valid:
        parser.add_argument('name', **kwargs)
    else:
        expected_err = 'Only one of the following parameters'
        with pytest.raises(ValueError, match=expected_err):
            parser.add_argument('name', **kwargs)


@pytest.mark.parametrize('kwargs', [({'choices_provider': fake_func}), ({'completer': fake_func})])
def test_apcustom_no_choices_callables_alongside_choices(kwargs) -> None:
    parser = Cmd2ArgumentParser()
    with pytest.raises(TypeError) as excinfo:
        parser.add_argument('name', choices=['my', 'choices', 'list'], **kwargs)
    assert 'None of the following parameters can be used alongside a choices parameter' in str(excinfo.value)


@pytest.mark.parametrize('kwargs', [({'choices_provider': fake_func}), ({'completer': fake_func})])
def test_apcustom_no_choices_callables_when_nargs_is_0(kwargs) -> None:
    parser = Cmd2ArgumentParser()
    with pytest.raises(TypeError) as excinfo:
        parser.add_argument('--name', action='store_true', **kwargs)
    assert 'None of the following parameters can be used on an action that takes no arguments' in str(excinfo.value)


def test_apcustom_usage() -> None:
    usage = "A custom usage statement"
    parser = Cmd2ArgumentParser(usage=usage)
    assert usage in parser.format_help()


def test_apcustom_nargs_help_format(cust_app) -> None:
    out, _err = run_cmd(cust_app, 'help range')
    assert 'Usage: range [-h] [--arg0 ARG0] [--arg1 ARG1{2}] [--arg2 ARG2{3+}]' in out[0]
    assert '             [--arg3 ARG3{2..3}]' in out[1]


def test_apcustom_nargs_range_validation(cust_app) -> None:
    # nargs = (3,)  # noqa: ERA001
    _out, err = run_cmd(cust_app, 'range --arg2 one two')
    assert 'Error: argument --arg2: expected at least 3 arguments' in err[2]

    _out, err = run_cmd(cust_app, 'range --arg2 one two three')
    assert not err

    _out, err = run_cmd(cust_app, 'range --arg2 one two three four')
    assert not err

    # nargs = (2,3)  # noqa: ERA001
    _out, err = run_cmd(cust_app, 'range --arg3 one')
    assert 'Error: argument --arg3: expected 2 to 3 arguments' in err[2]

    _out, err = run_cmd(cust_app, 'range --arg3 one two')
    assert not err

    _out, err = run_cmd(cust_app, 'range --arg2 one two three')
    assert not err


@pytest.mark.parametrize(
    ('nargs', 'expected_parts'),
    [
        # arg{2}
        (
            2,
            [("arg", True), ("{2}", False)],
        ),
        # arg{2+}
        (
            (2,),
            [("arg", True), ("{2+}", False)],
        ),
        # arg{0..5}
        (
            (0, 5),
            [("arg", True), ("{0..5}", False)],
        ),
    ],
)
def test_rich_metavar_parts(
    nargs: int | tuple[int, int | float],
    expected_parts: list[tuple[str, bool]],
) -> None:
    """
    Test cmd2's override of _rich_metavar_parts which handles custom nargs formats.

    :param nargs: the arguments nargs value
    :param expected_parts: list to compare to _rich_metavar_parts's return value

                           Each element in this list is a 2-item tuple.
                           item 1: one part of the argument string outputted by _format_args
                           item 2: boolean stating whether rich-argparse should color this part
    """
    parser = Cmd2ArgumentParser()
    help_formatter = parser._get_formatter()

    action = parser.add_argument("arg", nargs=nargs)  # type: ignore[arg-type]
    default_metavar = help_formatter._get_default_metavar_for_positional(action)

    parts = help_formatter._rich_metavar_parts(action, default_metavar)
    assert list(parts) == expected_parts


@pytest.mark.parametrize(
    'nargs_tuple',
    [
        (),
        ('f', 5),
        (5, 'f'),
        (1, 2, 3),
    ],
)
def test_apcustom_narg_invalid_tuples(nargs_tuple) -> None:
    parser = Cmd2ArgumentParser()
    expected_err = 'Ranged values for nargs must be a tuple of 1 or 2 integers'
    with pytest.raises(ValueError, match=expected_err):
        parser.add_argument('invalid_tuple', nargs=nargs_tuple)


def test_apcustom_narg_tuple_order() -> None:
    parser = Cmd2ArgumentParser()
    expected_err = 'Invalid nargs range. The first value must be less than the second'
    with pytest.raises(ValueError, match=expected_err):
        parser.add_argument('invalid_tuple', nargs=(2, 1))


def test_apcustom_narg_tuple_negative() -> None:
    parser = Cmd2ArgumentParser()
    expected_err = 'Negative numbers are invalid for nargs range'
    with pytest.raises(ValueError, match=expected_err):
        parser.add_argument('invalid_tuple', nargs=(-1, 1))


def test_apcustom_narg_tuple_zero_base() -> None:
    parser = Cmd2ArgumentParser()
    arg = parser.add_argument('arg', nargs=(0,))
    assert arg.nargs == argparse.ZERO_OR_MORE
    assert arg.nargs_range is None
    assert "[arg ...]" in parser.format_help()

    parser = Cmd2ArgumentParser()
    arg = parser.add_argument('arg', nargs=(0, 1))
    assert arg.nargs == argparse.OPTIONAL
    assert arg.nargs_range is None
    assert "[arg]" in parser.format_help()

    parser = Cmd2ArgumentParser()
    arg = parser.add_argument('arg', nargs=(0, 3))
    assert arg.nargs == argparse.ZERO_OR_MORE
    assert arg.nargs_range == (0, 3)
    assert "arg{0..3}" in parser.format_help()


def test_apcustom_narg_tuple_one_base() -> None:
    parser = Cmd2ArgumentParser()
    arg = parser.add_argument('arg', nargs=(1,))
    assert arg.nargs == argparse.ONE_OR_MORE
    assert arg.nargs_range is None
    assert "arg [arg ...]" in parser.format_help()

    parser = Cmd2ArgumentParser()
    arg = parser.add_argument('arg', nargs=(1, 5))
    assert arg.nargs == argparse.ONE_OR_MORE
    assert arg.nargs_range == (1, 5)
    assert "arg{1..5}" in parser.format_help()


def test_apcustom_narg_tuple_other_ranges() -> None:
    # Test range with no upper bound on max
    parser = Cmd2ArgumentParser()
    arg = parser.add_argument('arg', nargs=(2,))
    assert arg.nargs == argparse.ONE_OR_MORE
    assert arg.nargs_range == (2, constants.INFINITY)

    # Test finite range
    parser = Cmd2ArgumentParser()
    arg = parser.add_argument('arg', nargs=(2, 5))
    assert arg.nargs == argparse.ONE_OR_MORE
    assert arg.nargs_range == (2, 5)


def test_apcustom_print_message(capsys) -> None:
    import sys

    test_message = 'The test message'

    # Specify the file
    parser = Cmd2ArgumentParser()
    parser._print_message(test_message, file=sys.stdout)
    out, err = capsys.readouterr()
    assert test_message in out

    # Make sure file defaults to sys.stderr
    parser = Cmd2ArgumentParser()
    parser._print_message(test_message)
    out, err = capsys.readouterr()
    assert test_message in err


def test_generate_range_error() -> None:
    # max is INFINITY
    err_str = generate_range_error(1, constants.INFINITY)
    assert err_str == "expected at least 1 argument"

    err_str = generate_range_error(2, constants.INFINITY)
    assert err_str == "expected at least 2 arguments"

    # min and max are equal
    err_str = generate_range_error(1, 1)
    assert err_str == "expected 1 argument"

    err_str = generate_range_error(2, 2)
    assert err_str == "expected 2 arguments"

    # min and max are not equal
    err_str = generate_range_error(0, 1)
    assert err_str == "expected 0 to 1 argument"

    err_str = generate_range_error(0, 2)
    assert err_str == "expected 0 to 2 arguments"


def test_apcustom_metavar_tuple() -> None:
    # Test the case when a tuple metavar is used with nargs an integer > 1
    parser = Cmd2ArgumentParser()
    parser.add_argument('--aflag', nargs=2, metavar=('foo', 'bar'), help='This is a test')
    assert '[--aflag foo bar]' in parser.format_help()


def test_cmd2_attribute_wrapper() -> None:
    initial_val = 5
    wrapper = cmd2.Cmd2AttributeWrapper(initial_val)
    assert wrapper.get() == initial_val

    new_val = 22
    wrapper.set(new_val)
    assert wrapper.get() == new_val


def test_completion_items_as_choices(capsys) -> None:
    """Test cmd2's patch to Argparse._check_value() which supports CompletionItems as choices.
    Choices are compared to CompletionItems.orig_value instead of the CompletionItem instance.
    """
    from cmd2.argparse_custom import (
        CompletionItem,
    )

    ##############################################################
    # Test CompletionItems with str values
    ##############################################################
    choices = [CompletionItem("1", "Description One"), CompletionItem("2", "Two")]
    parser = Cmd2ArgumentParser()
    parser.add_argument("choices_arg", type=str, choices=choices)

    # First test valid choices. Confirm the parsed data matches the correct type of str.
    args = parser.parse_args(['1'])
    assert args.choices_arg == '1'

    args = parser.parse_args(['2'])
    assert args.choices_arg == '2'

    # Next test invalid choice
    with pytest.raises(SystemExit):
        args = parser.parse_args(['3'])

    # Confirm error text contains correct value type of str
    _out, err = capsys.readouterr()
    assert "invalid choice: '3' (choose from '1', '2')" in err

    ##############################################################
    # Test CompletionItems with int values
    ##############################################################
    choices = [CompletionItem(1, "Description One"), CompletionItem(2, "Two")]
    parser = Cmd2ArgumentParser()
    parser.add_argument("choices_arg", type=int, choices=choices)

    # First test valid choices. Confirm the parsed data matches the correct type of int.
    args = parser.parse_args(['1'])
    assert args.choices_arg == 1

    args = parser.parse_args(['2'])
    assert args.choices_arg == 2

    # Next test invalid choice
    with pytest.raises(SystemExit):
        args = parser.parse_args(['3'])

    # Confirm error text contains correct value type of int
    _out, err = capsys.readouterr()
    assert 'invalid choice: 3 (choose from 1, 2)' in err



"""Cmd2 unit/functional testing"""

import argparse
import sys
from collections.abc import Callable
from contextlib import redirect_stderr
from typing import (
    TYPE_CHECKING,
    ParamSpec,
    TextIO,
    TypeVar,
    cast,
)
from unittest import mock

import pytest

import cmd2
from cmd2 import rich_utils as ru
from cmd2.rl_utils import readline
from cmd2.utils import StdSim

# For type hinting decorators
P = ParamSpec('P')
T = TypeVar('T')


def verify_help_text(cmd2_app: cmd2.Cmd, help_output: str | list[str], verbose_strings: list[str] | None = None) -> None:
    """This function verifies that all expected commands are present in the help text.

    :param cmd2_app: instance of cmd2.Cmd
    :param help_output: output of help, either as a string or list of strings
    :param verbose_strings: optional list of verbose strings to search for
    """
    help_text = help_output if isinstance(help_output, str) else ''.join(help_output)
    commands = cmd2_app.get_visible_commands()
    for command in commands:
        assert command in help_text

    if verbose_strings:
        for verbose_string in verbose_strings:
            assert verbose_string in help_text


# Output from the shortcuts command with default built-in shortcuts
SHORTCUTS_TXT = """Shortcuts for other commands:
!: shell
?: help
@: run_script
@@: _relative_run_script
"""


def normalize(block: str) -> list[str]:
    """Normalize a block of text to perform comparison.

    Strip newlines from the very beginning and very end  Then split into separate lines and strip trailing whitespace
    from each line.
    """
    assert isinstance(block, str)
    block = block.strip('\n')
    return [line.rstrip() for line in block.splitlines()]


def run_cmd(app: cmd2.Cmd, cmd: str) -> tuple[list[str], list[str]]:
    """Clear out and err StdSim buffers, run the command, and return out and err"""

    # Only capture sys.stdout if it's the same stream as self.stdout
    stdouts_match = app.stdout == sys.stdout

    # This will be used to capture app.stdout and sys.stdout
    copy_cmd_stdout = StdSim(cast(TextIO, app.stdout))

    # This will be used to capture sys.stderr
    copy_stderr = StdSim(sys.stderr)

    try:
        app.stdout = cast(TextIO, copy_cmd_stdout)
        if stdouts_match:
            sys.stdout = app.stdout
        with redirect_stderr(cast(TextIO, copy_stderr)):
            app.onecmd_plus_hooks(cmd)
    finally:
        app.stdout = cast(TextIO, copy_cmd_stdout.inner_stream)
        if stdouts_match:
            sys.stdout = app.stdout

    out = copy_cmd_stdout.getvalue()
    err = copy_stderr.getvalue()
    return normalize(out), normalize(err)


@pytest.fixture
def base_app() -> cmd2.Cmd:
    return cmd2.Cmd(include_py=True, include_ipy=True)


def with_ansi_style(style: ru.AllowStyle) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator which sets ru.ALLOW_STYLE before a function runs and restores it when it's done."""

    def arg_decorator(func: Callable[P, T]) -> Callable[P, T]:
        import functools

        @functools.wraps(func)
        def cmd_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            old = ru.ALLOW_STYLE
            ru.ALLOW_STYLE = style
            try:
                retval = func(*args, **kwargs)
            finally:
                ru.ALLOW_STYLE = old
            return retval

        return cmd_wrapper

    return arg_decorator


# These are odd file names for testing quoting of them
odd_file_names = ['nothingweird', 'has   spaces', '"is_double_quoted"', "'is_single_quoted'"]


def complete_tester(text: str, line: str, begidx: int, endidx: int, app: cmd2.Cmd) -> str | None:
    """This is a convenience function to test cmd2.complete() since
    in a unit test environment there is no actual console readline
    is monitoring. Therefore we use mock to provide readline data
    to complete().

    :param text: the string prefix we are attempting to match
    :param line: the current input line with leading whitespace removed
    :param begidx: the beginning index of the prefix text
    :param endidx: the ending index of the prefix text
    :param app: the cmd2 app that will run completions
    :return: The first matched string or None if there are no matches
             Matches are stored in app.completion_matches
             These matches also have been sorted by complete()
    """

    def get_line() -> str:
        return line

    def get_begidx() -> int:
        return begidx

    def get_endidx() -> int:
        return endidx

    # Run the readline tab completion function with readline mocks in place
    with (
        mock.patch.object(readline, 'get_line_buffer', get_line),
        mock.patch.object(readline, 'get_begidx', get_begidx),
        mock.patch.object(readline, 'get_endidx', get_endidx),
    ):
        return app.complete(text, 0)


def find_subcommand(action: argparse.ArgumentParser, subcmd_names: list[str]) -> argparse.ArgumentParser:
    if not subcmd_names:
        return action
    cur_subcmd = subcmd_names.pop(0)
    for sub_action in action._actions:
        if isinstance(sub_action, argparse._SubParsersAction):
            for choice_name, choice in sub_action.choices.items():
                if choice_name == cur_subcmd:
                    return find_subcommand(choice, subcmd_names)
            break
    raise ValueError(f"Could not find subcommand '{subcmd_names}'")


if TYPE_CHECKING:
    _Base = cmd2.Cmd
else:
    _Base = object


class ExternalTestMixin(_Base):
    """A cmd2 plugin (mixin class) that exposes an interface to execute application commands from python"""

    def __init__(self, *args, **kwargs):
        """Initializes the ExternalTestMixin.

        This class is intended to be used in multiple inheritance alongside `cmd2.Cmd` for an application class.
        When doing this multiple inheritance, it is imperative that this mixin class come first.

        :type self: cmd2.Cmd
        :param args: arguments to pass to the superclass
        :param kwargs: keyword arguments to pass to the superclass
        """
        # code placed here runs before cmd2 initializes
        super().__init__(*args, **kwargs)
        if not isinstance(self, cmd2.Cmd):
            raise TypeError('The ExternalTestMixin class is intended to be used in multiple inheritance with cmd2.Cmd')
        # code placed here runs after cmd2 initializes
        self._pybridge = cmd2.py_bridge.PyBridge(self)

    def app_cmd(self, command: str, echo: bool | None = None) -> cmd2.CommandResult:
        """
        Run the application command

        :param command: The application command as it would be written on the cmd2 application prompt
        :param echo: Flag whether the command's output should be echoed to stdout/stderr
        :return: A CommandResult object that captures stdout, stderr, and the command's result object
        """
        try:
            self._in_py = True
            return self._pybridge(command, echo=echo)

        finally:
            self._in_py = False

    def fixture_setup(self):
        """Replicates the behavior of `cmdloop()` to prepare the application state for testing.

        This method runs all preloop hooks and the preloop method to ensure the
        application is in the correct state before running a test.

        :type self: cmd2.Cmd
        """

        for func in self._preloop_hooks:
            func()
        self.preloop()

    def fixture_teardown(self):
        """Replicates the behavior of `cmdloop()` to tear down the application after a test.

        This method runs all postloop hooks and the postloop method to clean up
        the application state and ensure test isolation.

        :type self: cmd2.Cmd
        """
        for func in self._postloop_hooks:
            func()
        self.postloop()


class WithCommandSets(ExternalTestMixin, cmd2.Cmd):
    """Class for testing custom help_* methods which override docstring help."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


@pytest.fixture
def autoload_command_sets_app():
    return WithCommandSets(auto_load_commands=True)


@pytest.fixture
def manual_command_sets_app():
    return WithCommandSets(auto_load_commands=False)



#!/usr/bin/env python
"""A simple example demonstrating an application that asynchronously prints alerts, updates the prompt
and changes the window title.
"""

import random
import threading
import time

import cmd2
from cmd2 import (
    Color,
    stylize,
)

ALERTS = [
    "Watch as this application prints alerts and updates the prompt",
    "This will only happen when the prompt is present",
    "Notice how it doesn't interfere with your typing or cursor location",
    "Go ahead and type some stuff and move the cursor throughout the line",
    "Keep typing...",
    "Move that cursor...",
    "Pretty seamless, eh?",
    "Feedback can also be given in the window title. Notice the alert count up there?",
    "You can stop and start the alerts by typing stop_alerts and start_alerts",
    "This demo will now continue to print alerts at random intervals",
]


class AlerterApp(cmd2.Cmd):
    """An app that shows off async_alert() and async_update_prompt()."""

    def __init__(self, *args, **kwargs) -> None:
        """Initializer."""
        super().__init__(*args, **kwargs)

        self.prompt = "(APR)> "

        # The thread that will asynchronously alert the user of events
        self._stop_event = threading.Event()
        self._alerter_thread = threading.Thread()
        self._alert_count = 0
        self._next_alert_time = 0

        # Create some hooks to handle the starting and stopping of our thread
        self.register_preloop_hook(self._preloop_hook)
        self.register_postloop_hook(self._postloop_hook)

    def _preloop_hook(self) -> None:
        """Start the alerter thread."""
        # This runs after cmdloop() acquires self.terminal_lock, which will be locked until the prompt appears.
        # Therefore this is the best place to start the alerter thread since there is no risk of it alerting
        # before the prompt is displayed. You can also start it via a command if its not something that should
        # be running during the entire application. See do_start_alerts().
        self._stop_event.clear()

        self._alerter_thread = threading.Thread(name='alerter', target=self._alerter_thread_func)
        self._alerter_thread.start()

    def _postloop_hook(self) -> None:
        """Stops the alerter thread."""
        # After this function returns, cmdloop() releases self.terminal_lock which could make the alerter
        # thread think the prompt is on screen. Therefore this is the best place to stop the alerter thread.
        # You can also stop it via a command. See do_stop_alerts().
        self._stop_event.set()
        if self._alerter_thread.is_alive():
            self._alerter_thread.join()

    def do_start_alerts(self, _) -> None:
        """Starts the alerter thread."""
        if self._alerter_thread.is_alive():
            print("The alert thread is already started")
        else:
            self._stop_event.clear()
            self._alerter_thread = threading.Thread(name='alerter', target=self._alerter_thread_func)
            self._alerter_thread.start()

    def do_stop_alerts(self, _) -> None:
        """Stops the alerter thread."""
        self._stop_event.set()
        if self._alerter_thread.is_alive():
            self._alerter_thread.join()
        else:
            print("The alert thread is already stopped")

    def _get_alerts(self) -> list[str]:
        """Reports alerts
        :return: the list of alerts.
        """
        cur_time = time.monotonic()
        if cur_time < self._next_alert_time:
            return []

        alerts = []

        if self._alert_count < len(ALERTS):
            alerts.append(ALERTS[self._alert_count])
            self._alert_count += 1
            self._next_alert_time = cur_time + 4

        else:
            rand_num = random.randint(1, 20)
            if rand_num > 2:
                return []

            for _ in range(rand_num):
                self._alert_count += 1
                alerts.append(f"Alert {self._alert_count}")

            self._next_alert_time = 0

        return alerts

    def _generate_alert_str(self) -> str:
        """Combines alerts into one string that can be printed to the terminal
        :return: the alert string.
        """
        alert_str = ''
        alerts = self._get_alerts()

        longest_alert = max(ALERTS, key=len)
        num_asterisks = len(longest_alert) + 8

        for i, cur_alert in enumerate(alerts):
            # Use padding to center the alert
            padding = ' ' * int((num_asterisks - len(cur_alert)) / 2)

            if i > 0:
                alert_str += '\n'
            alert_str += '*' * num_asterisks + '\n'
            alert_str += padding + cur_alert + padding + '\n'
            alert_str += '*' * num_asterisks + '\n'

        return alert_str

    def _generate_colored_prompt(self) -> str:
        """Randomly generates a colored prompt
        :return: the new prompt.
        """
        rand_num = random.randint(1, 20)

        status_color = Color.DEFAULT

        if rand_num == 1:
            status_color = Color.BRIGHT_RED
        elif rand_num == 2:
            status_color = Color.BRIGHT_YELLOW
        elif rand_num == 3:
            status_color = Color.CYAN
        elif rand_num == 4:
            status_color = Color.BRIGHT_GREEN
        elif rand_num == 5:
            status_color = Color.BRIGHT_BLUE

        return stylize(self.visible_prompt, style=status_color)

    def _alerter_thread_func(self) -> None:
        """Prints alerts and updates the prompt any time the prompt is showing."""
        self._alert_count = 0
        self._next_alert_time = 0

        while not self._stop_event.is_set():
            # Always acquire terminal_lock before printing alerts or updating the prompt.
            # To keep the app responsive, do not block on this call.
            if self.terminal_lock.acquire(blocking=False):
                # Get any alerts that need to be printed
                alert_str = self._generate_alert_str()

                # Generate a new prompt
                new_prompt = self._generate_colored_prompt()

                # Check if we have alerts to print
                if alert_str:
                    # new_prompt is an optional parameter to async_alert()
                    self.async_alert(alert_str, new_prompt)
                    new_title = f"Alerts Printed: {self._alert_count}"
                    self.set_window_title(new_title)

                # Otherwise check if the prompt needs to be updated or refreshed
                elif self.prompt != new_prompt:
                    self.async_update_prompt(new_prompt)

                elif self.need_prompt_refresh():
                    self.async_refresh_prompt()

                # Don't forget to release the lock
                self.terminal_lock.release()

            self._stop_event.wait(0.5)


if __name__ == '__main__':
    import sys

    app = AlerterApp()
    app.set_window_title("Asynchronous Printer Test")
    sys.exit(app.cmdloop())



#!/usr/bin/env python
"""A sample application for cmd2.

Thanks to cmd2's built-in transcript testing capability, it also serves as a
test suite for transcript_example.py when used with the transcript_regex.txt transcript.

Running `python transcript_example.py -t transcript_regex.txt` will run all the commands in
the transcript against transcript_example.py, verifying that the output produced matches
the transcript.
"""

import random

import cmd2


class CmdLineApp(cmd2.Cmd):
    """Example cmd2 application."""

    # Setting this true makes it run a shell command if a cmd2/cmd command doesn't exist
    # default_to_shell = True  # noqa: ERA001
    MUMBLES = ('like', '...', 'um', 'er', 'hmmm', 'ahh')
    MUMBLE_FIRST = ('so', 'like', 'well')
    MUMBLE_LAST = ('right?',)

    def __init__(self) -> None:
        shortcuts = cmd2.DEFAULT_SHORTCUTS
        shortcuts.update({'&': 'speak'})
        super().__init__(multiline_commands=['orate'], shortcuts=shortcuts)

        # Make maxrepeats settable at runtime
        self.maxrepeats = 3
        self.add_settable(cmd2.Settable('maxrepeats', int, 'max repetitions for speak command', self))

    speak_parser = cmd2.Cmd2ArgumentParser()
    speak_parser.add_argument('-p', '--piglatin', action='store_true', help='atinLay')
    speak_parser.add_argument('-s', '--shout', action='store_true', help='N00B EMULATION MODE')
    speak_parser.add_argument('-r', '--repeat', type=int, help='output [n] times')
    speak_parser.add_argument('words', nargs='+', help='words to say')

    @cmd2.with_argparser(speak_parser)
    def do_speak(self, args) -> None:
        """Repeats what you tell me to."""
        words = []
        for word in args.words:
            if args.piglatin:
                word = f'{word[1:]}{word[0]}ay'
            if args.shout:
                word = word.upper()
            words.append(word)
        repetitions = args.repeat or 1
        for _ in range(min(repetitions, self.maxrepeats)):
            # .poutput handles newlines, and accommodates output redirection too
            self.poutput(' '.join(words))

    do_say = do_speak  # now "say" is a synonym for "speak"
    do_orate = do_speak  # another synonym, but this one takes multi-line input

    mumble_parser = cmd2.Cmd2ArgumentParser()
    mumble_parser.add_argument('-r', '--repeat', type=int, help='how many times to repeat')
    mumble_parser.add_argument('words', nargs='+', help='words to say')

    @cmd2.with_argparser(mumble_parser)
    def do_mumble(self, args) -> None:
        """Mumbles what you tell me to."""
        repetitions = args.repeat or 1
        for _ in range(min(repetitions, self.maxrepeats)):
            output = []
            if random.random() < 0.33:
                output.append(random.choice(self.MUMBLE_FIRST))
            for word in args.words:
                if random.random() < 0.40:
                    output.append(random.choice(self.MUMBLES))
                output.append(word)
            if random.random() < 0.25:
                output.append(random.choice(self.MUMBLE_LAST))
            self.poutput(' '.join(output))


if __name__ == '__main__':
    import sys

    c = CmdLineApp()
    sys.exit(c.cmdloop())



"""Test plugin infrastructure and hooks."""

import argparse
import sys
from typing import NoReturn
from unittest import (
    mock,
)

import pytest

import cmd2
from cmd2 import (
    Cmd2ArgumentParser,
    exceptions,
    plugin,
    with_argparser,
)


class Plugin:
    """A mixin class for testing hook registration and calling"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.reset_counters()

    def reset_counters(self) -> None:
        self.called_preparse = 0
        self.called_postparsing = 0
        self.called_precmd = 0
        self.called_postcmd = 0
        self.called_cmdfinalization = 0

    ###
    #
    # preloop and postloop hooks
    # which share the same signature and are thus interchangeable
    #
    ###
    def prepost_hook_one(self) -> None:
        """Method used for preloop or postloop hooks"""
        self.poutput("one")

    def prepost_hook_two(self) -> None:
        """Another method used for preloop or postloop hooks"""
        self.poutput("two")

    def prepost_hook_too_many_parameters(self, param) -> None:
        """A preloop or postloop hook with too many parameters"""

    def prepost_hook_with_wrong_return_annotation(self) -> bool:
        """A preloop or postloop hook with incorrect return type"""

    ###
    #
    # preparse hook
    #
    ###
    def preparse(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """Preparsing hook"""
        self.called_preparse += 1
        return data

    ###
    #
    # Postparsing hooks
    #
    ###
    def postparse_hook(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """A postparsing hook"""
        self.called_postparsing += 1
        return data

    def postparse_hook_stop(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """A postparsing hook with requests application exit"""
        self.called_postparsing += 1
        data.stop = True
        return data

    def postparse_hook_emptystatement(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """A postparsing hook with raises an EmptyStatement exception"""
        self.called_postparsing += 1
        raise exceptions.EmptyStatement

    def postparse_hook_exception(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """A postparsing hook which raises an exception"""
        self.called_postparsing += 1
        raise ValueError

    def postparse_hook_too_many_parameters(self, data1, data2) -> cmd2.plugin.PostparsingData:
        """A postparsing hook with too many parameters"""

    def postparse_hook_undeclared_parameter_annotation(self, data) -> cmd2.plugin.PostparsingData:
        """A postparsing hook with an undeclared parameter type"""

    def postparse_hook_wrong_parameter_annotation(self, data: str) -> cmd2.plugin.PostparsingData:
        """A postparsing hook with the wrong parameter type"""

    def postparse_hook_undeclared_return_annotation(self, data: cmd2.plugin.PostparsingData) -> None:
        """A postparsing hook with an undeclared return type"""

    def postparse_hook_wrong_return_annotation(self, data: cmd2.plugin.PostparsingData) -> str:
        """A postparsing hook with the wrong return type"""

    ###
    #
    # precommand hooks, some valid, some invalid
    #
    ###
    def precmd(self, statement: cmd2.Statement) -> cmd2.Statement:
        """Override cmd.Cmd method"""
        self.called_precmd += 1
        return statement

    def precmd_hook(self, data: plugin.PrecommandData) -> plugin.PrecommandData:
        """A precommand hook"""
        self.called_precmd += 1
        return data

    def precmd_hook_emptystatement(self, data: plugin.PrecommandData) -> plugin.PrecommandData:
        """A precommand hook which raises an EmptyStatement exception"""
        self.called_precmd += 1
        raise exceptions.EmptyStatement

    def precmd_hook_exception(self, data: plugin.PrecommandData) -> plugin.PrecommandData:
        """A precommand hook which raises an exception"""
        self.called_precmd += 1
        raise ValueError

    def precmd_hook_not_enough_parameters(self) -> plugin.PrecommandData:
        """A precommand hook with no parameters"""

    def precmd_hook_too_many_parameters(self, one: plugin.PrecommandData, two: str) -> plugin.PrecommandData:
        """A precommand hook with too many parameters"""
        return one

    def precmd_hook_no_parameter_annotation(self, data) -> plugin.PrecommandData:
        """A precommand hook with no type annotation on the parameter"""
        return data

    def precmd_hook_wrong_parameter_annotation(self, data: str) -> plugin.PrecommandData:
        """A precommand hook with the incorrect type annotation on the parameter"""
        return data

    def precmd_hook_no_return_annotation(self, data: plugin.PrecommandData):
        """A precommand hook with no type annotation on the return value"""
        return data

    def precmd_hook_wrong_return_annotation(self, data: plugin.PrecommandData) -> cmd2.Statement:
        return self.statement_parser.parse('hi there')

    ###
    #
    # postcommand hooks, some valid, some invalid
    #
    ###
    def postcmd(self, stop: bool, statement: cmd2.Statement) -> bool:
        """Override cmd.Cmd method"""
        self.called_postcmd += 1
        return stop

    def postcmd_hook(self, data: plugin.PostcommandData) -> plugin.PostcommandData:
        """A postcommand hook"""
        self.called_postcmd += 1
        return data

    def postcmd_hook_exception(self, data: plugin.PostcommandData) -> plugin.PostcommandData:
        """A postcommand hook with raises an exception"""
        self.called_postcmd += 1
        raise ZeroDivisionError

    def postcmd_hook_not_enough_parameters(self) -> plugin.PostcommandData:
        """A precommand hook with no parameters"""

    def postcmd_hook_too_many_parameters(self, one: plugin.PostcommandData, two: str) -> plugin.PostcommandData:
        """A precommand hook with too many parameters"""
        return one

    def postcmd_hook_no_parameter_annotation(self, data) -> plugin.PostcommandData:
        """A precommand hook with no type annotation on the parameter"""
        return data

    def postcmd_hook_wrong_parameter_annotation(self, data: str) -> plugin.PostcommandData:
        """A precommand hook with the incorrect type annotation on the parameter"""
        return data

    def postcmd_hook_no_return_annotation(self, data: plugin.PostcommandData):
        """A precommand hook with no type annotation on the return value"""
        return data

    def postcmd_hook_wrong_return_annotation(self, data: plugin.PostcommandData) -> cmd2.Statement:
        return self.statement_parser.parse('hi there')

    ###
    #
    # command finalization hooks, some valid, some invalid
    #
    ###
    def cmdfinalization_hook(self, data: plugin.CommandFinalizationData) -> plugin.CommandFinalizationData:
        """A command finalization hook."""
        self.called_cmdfinalization += 1
        return data

    def cmdfinalization_hook_stop(self, data: cmd2.plugin.CommandFinalizationData) -> cmd2.plugin.CommandFinalizationData:
        """A command finalization hook which requests application exit"""
        self.called_cmdfinalization += 1
        data.stop = True
        return data

    def cmdfinalization_hook_exception(self, data: cmd2.plugin.CommandFinalizationData) -> cmd2.plugin.CommandFinalizationData:
        """A command finalization hook which raises an exception"""
        self.called_cmdfinalization += 1
        raise ValueError

    def cmdfinalization_hook_system_exit(
        self, data: cmd2.plugin.CommandFinalizationData
    ) -> cmd2.plugin.CommandFinalizationData:
        """A command finalization hook which raises a SystemExit"""
        self.called_cmdfinalization += 1
        raise SystemExit(5)

    def cmdfinalization_hook_keyboard_interrupt(
        self, data: cmd2.plugin.CommandFinalizationData
    ) -> cmd2.plugin.CommandFinalizationData:
        """A command finalization hook which raises a KeyboardInterrupt"""
        self.called_cmdfinalization += 1
        raise KeyboardInterrupt

    def cmdfinalization_hook_passthrough_exception(
        self, data: cmd2.plugin.CommandFinalizationData
    ) -> cmd2.plugin.CommandFinalizationData:
        """A command finalization hook which raises a PassThroughException"""
        self.called_cmdfinalization += 1
        wrapped_ex = OSError("Pass me up")
        raise exceptions.PassThroughException(wrapped_ex=wrapped_ex)

    def cmdfinalization_hook_not_enough_parameters(self) -> plugin.CommandFinalizationData:
        """A command finalization hook with no parameters."""

    def cmdfinalization_hook_too_many_parameters(
        self, one: plugin.CommandFinalizationData, two: str
    ) -> plugin.CommandFinalizationData:
        """A command finalization hook with too many parameters."""
        return one

    def cmdfinalization_hook_no_parameter_annotation(self, data) -> plugin.CommandFinalizationData:
        """A command finalization hook with no type annotation on the parameter."""
        return data

    def cmdfinalization_hook_wrong_parameter_annotation(self, data: str) -> plugin.CommandFinalizationData:
        """A command finalization hook with the incorrect type annotation on the parameter."""
        return data

    def cmdfinalization_hook_no_return_annotation(self, data: plugin.CommandFinalizationData):
        """A command finalizationhook with no type annotation on the return value."""
        return data

    def cmdfinalization_hook_wrong_return_annotation(self, data: plugin.CommandFinalizationData) -> cmd2.Statement:
        """A command finalization hook with the wrong return type annotation."""
        return self.statement_parser.parse('hi there')


class PluggedApp(Plugin, cmd2.Cmd):
    """A sample app with a plugin mixed in"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def do_say(self, statement) -> None:
        """Repeat back the arguments"""
        self.poutput(statement)

    def do_skip_postcmd_hooks(self, _) -> NoReturn:
        self.poutput("In do_skip_postcmd_hooks")
        raise exceptions.SkipPostcommandHooks

    parser = Cmd2ArgumentParser(description="Test parser")
    parser.add_argument("my_arg", help="some help text")

    @with_argparser(parser)
    def do_argparse_cmd(self, namespace: argparse.Namespace) -> None:
        """Repeat back the arguments"""
        self.poutput(namespace.cmd2_statement.get())


###
#
# test pre and postloop hooks
#
###
def test_register_preloop_hook_too_many_parameters() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_preloop_hook(app.prepost_hook_too_many_parameters)


def test_register_preloop_hook_with_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_preloop_hook(app.prepost_hook_with_wrong_return_annotation)


def test_preloop_hook(capsys) -> None:
    # Need to patch sys.argv so cmd2 doesn't think it was called with arguments equal to the py.test args
    testargs = ["prog", "say hello", 'quit']

    with mock.patch.object(sys, 'argv', testargs):
        app = PluggedApp()

    app.register_preloop_hook(app.prepost_hook_one)
    app.cmdloop()
    out, err = capsys.readouterr()
    assert out == 'one\nhello\n'
    assert not err


def test_preloop_hooks(capsys) -> None:
    # Need to patch sys.argv so cmd2 doesn't think it was called with arguments equal to the py.test args
    testargs = ["prog", "say hello", 'quit']

    with mock.patch.object(sys, 'argv', testargs):
        app = PluggedApp()

    app.register_preloop_hook(app.prepost_hook_one)
    app.register_preloop_hook(app.prepost_hook_two)
    app.cmdloop()
    out, err = capsys.readouterr()
    assert out == 'one\ntwo\nhello\n'
    assert not err


def test_register_postloop_hook_too_many_parameters() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postloop_hook(app.prepost_hook_too_many_parameters)


def test_register_postloop_hook_with_wrong_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postloop_hook(app.prepost_hook_with_wrong_return_annotation)


def test_postloop_hook(capsys) -> None:
    # Need to patch sys.argv so cmd2 doesn't think it was called with arguments equal to the py.test args
    testargs = ["prog", "say hello", 'quit']

    with mock.patch.object(sys, 'argv', testargs):
        app = PluggedApp()

    app.register_postloop_hook(app.prepost_hook_one)
    app.cmdloop()
    out, err = capsys.readouterr()
    assert out == 'hello\none\n'
    assert not err


def test_postloop_hooks(capsys) -> None:
    # Need to patch sys.argv so cmd2 doesn't think it was called with arguments equal to the py.test args
    testargs = ["prog", "say hello", 'quit']

    with mock.patch.object(sys, 'argv', testargs):
        app = PluggedApp()

    app.register_postloop_hook(app.prepost_hook_one)
    app.register_postloop_hook(app.prepost_hook_two)
    app.cmdloop()
    out, err = capsys.readouterr()
    assert out == 'hello\none\ntwo\n'
    assert not err


###
#
# test preparse hook
#
###
def test_preparse(capsys) -> None:
    app = PluggedApp()
    app.register_postparsing_hook(app.preparse)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert app.called_preparse == 1


###
#
# test postparsing hooks
#
###
def test_postparsing_hook_too_many_parameters() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postparsing_hook(app.postparse_hook_too_many_parameters)


def test_postparsing_hook_undeclared_parameter_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postparsing_hook(app.postparse_hook_undeclared_parameter_annotation)


def test_postparsing_hook_wrong_parameter_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postparsing_hook(app.postparse_hook_wrong_parameter_annotation)


def test_postparsing_hook_undeclared_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postparsing_hook(app.postparse_hook_undeclared_return_annotation)


def test_postparsing_hook_wrong_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postparsing_hook(app.postparse_hook_wrong_return_annotation)


def test_postparsing_hook(capsys) -> None:
    app = PluggedApp()
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert not app.called_postparsing

    app.reset_counters()
    app.register_postparsing_hook(app.postparse_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert app.called_postparsing == 1

    # register the function again, so it should be called twice
    app.reset_counters()
    app.register_postparsing_hook(app.postparse_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert app.called_postparsing == 2


def test_postparsing_hook_stop_first(capsys) -> None:
    app = PluggedApp()
    app.register_postparsing_hook(app.postparse_hook_stop)
    stop = app.onecmd_plus_hooks('say hello')
    assert app.called_postparsing == 1
    assert stop

    # register another function but it shouldn't be called
    app.reset_counters()
    app.register_postparsing_hook(app.postparse_hook)
    stop = app.onecmd_plus_hooks('say hello')
    assert app.called_postparsing == 1
    assert stop


def test_postparsing_hook_stop_second(capsys) -> None:
    app = PluggedApp()
    app.register_postparsing_hook(app.postparse_hook)
    stop = app.onecmd_plus_hooks('say hello')
    assert app.called_postparsing == 1
    assert not stop

    # register another function and make sure it gets called
    app.reset_counters()
    app.register_postparsing_hook(app.postparse_hook_stop)
    stop = app.onecmd_plus_hooks('say hello')
    assert app.called_postparsing == 2
    assert stop

    # register a third function which shouldn't be called
    app.reset_counters()
    app.register_postparsing_hook(app.postparse_hook)
    stop = app.onecmd_plus_hooks('say hello')
    assert app.called_postparsing == 2
    assert stop


def test_postparsing_hook_emptystatement_first(capsys) -> None:
    app = PluggedApp()
    app.register_postparsing_hook(app.postparse_hook_emptystatement)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert not err
    assert app.called_postparsing == 1

    # register another function but it shouldn't be called
    app.reset_counters()
    stop = app.register_postparsing_hook(app.postparse_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert not err
    assert app.called_postparsing == 1


def test_postparsing_hook_emptystatement_second(capsys) -> None:
    app = PluggedApp()
    app.register_postparsing_hook(app.postparse_hook)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert out == 'hello\n'
    assert not err
    assert app.called_postparsing == 1

    # register another function and make sure it gets called
    app.reset_counters()
    app.register_postparsing_hook(app.postparse_hook_emptystatement)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert not err
    assert app.called_postparsing == 2

    # register a third function which shouldn't be called
    app.reset_counters()
    app.register_postparsing_hook(app.postparse_hook)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert not err
    assert app.called_postparsing == 2


def test_postparsing_hook_exception(capsys) -> None:
    app = PluggedApp()
    app.register_postparsing_hook(app.postparse_hook_exception)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert err
    assert app.called_postparsing == 1

    # register another function, but it shouldn't be called
    app.reset_counters()
    app.register_postparsing_hook(app.postparse_hook)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert err
    assert app.called_postparsing == 1


###
#
# test precmd hooks
#
#####
def test_register_precmd_hook_parameter_count() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_precmd_hook(app.precmd_hook_not_enough_parameters)
    with pytest.raises(TypeError):
        app.register_precmd_hook(app.precmd_hook_too_many_parameters)


def test_register_precmd_hook_no_parameter_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_precmd_hook(app.precmd_hook_no_parameter_annotation)


def test_register_precmd_hook_wrong_parameter_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_precmd_hook(app.precmd_hook_wrong_parameter_annotation)


def test_register_precmd_hook_no_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_precmd_hook(app.precmd_hook_no_return_annotation)


def test_register_precmd_hook_wrong_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_precmd_hook(app.precmd_hook_wrong_return_annotation)


def test_precmd_hook(capsys) -> None:
    app = PluggedApp()
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    # without registering any hooks, precmd() should be called
    assert app.called_precmd == 1

    app.reset_counters()
    app.register_precmd_hook(app.precmd_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    # with one hook registered, we should get precmd() and the hook
    assert app.called_precmd == 2

    # register the function again, so it should be called twice
    app.reset_counters()
    app.register_precmd_hook(app.precmd_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    # with two hooks registered, we should get precmd() and both hooks
    assert app.called_precmd == 3


def test_precmd_hook_emptystatement_first(capsys) -> None:
    app = PluggedApp()
    app.register_precmd_hook(app.precmd_hook_emptystatement)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert not err
    # since the registered hooks are called before precmd(), if a registered
    # hook throws an exception, precmd() is never called
    assert app.called_precmd == 1

    # register another function but it shouldn't be called
    app.reset_counters()
    stop = app.register_precmd_hook(app.precmd_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert not err
    # the exception raised by the first hook should prevent the second
    # hook from being called, and it also prevents precmd() from being
    # called
    assert app.called_precmd == 1


def test_precmd_hook_emptystatement_second(capsys) -> None:
    app = PluggedApp()
    app.register_precmd_hook(app.precmd_hook)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert out == 'hello\n'
    assert not err
    # with one hook registered, we should get precmd() and the hook
    assert app.called_precmd == 2

    # register another function and make sure it gets called
    app.reset_counters()
    app.register_precmd_hook(app.precmd_hook_emptystatement)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert not err
    # since the registered hooks are called before precmd(), if a registered
    # hook throws an exception, precmd() is never called
    assert app.called_precmd == 2

    # register a third function which shouldn't be called
    app.reset_counters()
    app.register_precmd_hook(app.precmd_hook)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert not out
    assert not err
    # the exception raised by the second hook should prevent the third
    # hook from being called. since the registered hooks are called before precmd(),
    # if a registered hook throws an exception, precmd() is never called
    assert app.called_precmd == 2


###
#
# test postcmd hooks
#
####
def test_register_postcmd_hook_parameter_count() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postcmd_hook(app.postcmd_hook_not_enough_parameters)
    with pytest.raises(TypeError):
        app.register_postcmd_hook(app.postcmd_hook_too_many_parameters)


def test_register_postcmd_hook_no_parameter_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postcmd_hook(app.postcmd_hook_no_parameter_annotation)


def test_register_postcmd_hook_wrong_parameter_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postcmd_hook(app.postcmd_hook_wrong_parameter_annotation)


def test_register_postcmd_hook_no_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postcmd_hook(app.postcmd_hook_no_return_annotation)


def test_register_postcmd_hook_wrong_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_postcmd_hook(app.postcmd_hook_wrong_return_annotation)


def test_postcmd(capsys) -> None:
    app = PluggedApp()
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    # without registering any hooks, postcmd() should be called
    assert app.called_postcmd == 1

    app.reset_counters()
    app.register_postcmd_hook(app.postcmd_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    # with one hook registered, we should get precmd() and the hook
    assert app.called_postcmd == 2

    # register the function again, so it should be called twice
    app.reset_counters()
    app.register_postcmd_hook(app.postcmd_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    # with two hooks registered, we should get precmd() and both hooks
    assert app.called_postcmd == 3


def test_postcmd_exception_first(capsys) -> None:
    app = PluggedApp()
    app.register_postcmd_hook(app.postcmd_hook_exception)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert out == 'hello\n'
    assert err
    # since the registered hooks are called before postcmd(), if a registered
    # hook throws an exception, postcmd() is never called. So we should have
    # a count of one because we called the hook that raised the exception
    assert app.called_postcmd == 1

    # register another function but it shouldn't be called
    app.reset_counters()
    stop = app.register_postcmd_hook(app.postcmd_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert out == 'hello\n'
    assert err
    # the exception raised by the first hook should prevent the second
    # hook from being called, and it also prevents postcmd() from being
    # called
    assert app.called_postcmd == 1


def test_postcmd_exception_second(capsys) -> None:
    app = PluggedApp()
    app.register_postcmd_hook(app.postcmd_hook)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert out == 'hello\n'
    assert not err
    # with one hook registered, we should get the hook and postcmd()
    assert app.called_postcmd == 2

    # register another function which should be called
    app.reset_counters()
    stop = app.register_postcmd_hook(app.postcmd_hook_exception)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert out == 'hello\n'
    assert err
    # the exception raised by the first hook should prevent the second
    # hook from being called, and it also prevents postcmd() from being
    # called. So we have the first hook, and the second hook, which raised
    # the exception
    assert app.called_postcmd == 2


##
#
# command finalization
#
###
def test_register_cmdfinalization_hook_parameter_count() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_cmdfinalization_hook(app.cmdfinalization_hook_not_enough_parameters)
    with pytest.raises(TypeError):
        app.register_cmdfinalization_hook(app.cmdfinalization_hook_too_many_parameters)


def test_register_cmdfinalization_hook_no_parameter_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_cmdfinalization_hook(app.cmdfinalization_hook_no_parameter_annotation)


def test_register_cmdfinalization_hook_wrong_parameter_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_cmdfinalization_hook(app.cmdfinalization_hook_wrong_parameter_annotation)


def test_register_cmdfinalization_hook_no_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_cmdfinalization_hook(app.cmdfinalization_hook_no_return_annotation)


def test_register_cmdfinalization_hook_wrong_return_annotation() -> None:
    app = PluggedApp()
    with pytest.raises(TypeError):
        app.register_cmdfinalization_hook(app.cmdfinalization_hook_wrong_return_annotation)


def test_cmdfinalization(capsys) -> None:
    app = PluggedApp()
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert app.called_cmdfinalization == 0

    app.register_cmdfinalization_hook(app.cmdfinalization_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert app.called_cmdfinalization == 1

    # register the function again, so it should be called twice
    app.reset_counters()
    app.register_cmdfinalization_hook(app.cmdfinalization_hook)
    app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert app.called_cmdfinalization == 2


def test_cmdfinalization_stop_first(capsys) -> None:
    app = PluggedApp()
    app.register_cmdfinalization_hook(app.cmdfinalization_hook_stop)
    app.register_cmdfinalization_hook(app.cmdfinalization_hook)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert app.called_cmdfinalization == 2
    assert stop


def test_cmdfinalization_stop_second(capsys) -> None:
    app = PluggedApp()
    app.register_cmdfinalization_hook(app.cmdfinalization_hook)
    app.register_cmdfinalization_hook(app.cmdfinalization_hook_stop)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert out == 'hello\n'
    assert not err
    assert app.called_cmdfinalization == 2
    assert stop


def test_cmdfinalization_hook_exception(capsys) -> None:
    app = PluggedApp()
    app.register_cmdfinalization_hook(app.cmdfinalization_hook_exception)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert out == 'hello\n'
    assert err
    assert app.called_cmdfinalization == 1

    # register another function, but it shouldn't be called
    app.reset_counters()
    app.register_cmdfinalization_hook(app.cmdfinalization_hook)
    stop = app.onecmd_plus_hooks('say hello')
    out, err = capsys.readouterr()
    assert not stop
    assert out == 'hello\n'
    assert err
    assert app.called_cmdfinalization == 1


def test_cmdfinalization_hook_system_exit() -> None:
    app = PluggedApp()
    app.register_cmdfinalization_hook(app.cmdfinalization_hook_system_exit)
    stop = app.onecmd_plus_hooks('say hello')
    assert stop
    assert app.called_cmdfinalization == 1
    assert app.exit_code == 5


def test_cmdfinalization_hook_keyboard_interrupt() -> None:
    app = PluggedApp()
    app.register_cmdfinalization_hook(app.cmdfinalization_hook_keyboard_interrupt)

    # First make sure KeyboardInterrupt isn't raised unless told to
    stop = app.onecmd_plus_hooks('say hello', raise_keyboard_interrupt=False)
    assert not stop
    assert app.called_cmdfinalization == 1

    # Now enable raising the KeyboardInterrupt
    app.reset_counters()
    with pytest.raises(KeyboardInterrupt):
        stop = app.onecmd_plus_hooks('say hello', raise_keyboard_interrupt=True)
    assert not stop
    assert app.called_cmdfinalization == 1

    # Now make sure KeyboardInterrupt isn't raised if stop is already True
    app.reset_counters()
    stop = app.onecmd_plus_hooks('quit', raise_keyboard_interrupt=True)
    assert stop
    assert app.called_cmdfinalization == 1


def test_cmdfinalization_hook_passthrough_exception() -> None:
    app = PluggedApp()
    app.register_cmdfinalization_hook(app.cmdfinalization_hook_passthrough_exception)

    expected_err = "Pass me up"
    with pytest.raises(OSError, match=expected_err):
        app.onecmd_plus_hooks('say hello')
    assert app.called_cmdfinalization == 1


def test_skip_postcmd_hooks(capsys) -> None:
    app = PluggedApp()
    app.register_postcmd_hook(app.postcmd_hook)
    app.register_cmdfinalization_hook(app.cmdfinalization_hook)

    # Cause a SkipPostcommandHooks exception and verify no postcmd stuff runs but cmdfinalization_hook still does
    app.onecmd_plus_hooks('skip_postcmd_hooks')
    out, _err = capsys.readouterr()
    assert "In do_skip_postcmd_hooks" in out
    assert app.called_postcmd == 0
    assert app.called_cmdfinalization == 1


def test_cmd2_argparse_exception(capsys) -> None:
    """Verify Cmd2ArgparseErrors raised after calling a command prevent postcmd events from
    running but do not affect cmdfinalization events
    """
    app = PluggedApp()
    app.register_postcmd_hook(app.postcmd_hook)
    app.register_cmdfinalization_hook(app.cmdfinalization_hook)

    # First generate no exception and make sure postcmd_hook, postcmd, and cmdfinalization_hook run
    app.onecmd_plus_hooks('argparse_cmd arg_val')
    out, err = capsys.readouterr()
    assert out == 'arg_val\n'
    assert not err
    assert app.called_postcmd == 2
    assert app.called_cmdfinalization == 1

    app.reset_counters()

    # Next cause an argparse exception and verify no postcmd stuff runs but cmdfinalization_hook still does
    app.onecmd_plus_hooks('argparse_cmd')
    out, err = capsys.readouterr()
    assert not out
    assert "Error: the following arguments are required: my_arg" in err
    assert app.called_postcmd == 0
    assert app.called_cmdfinalization == 1



#!/usr/bin/env python
"""A simple example demonstrating calling an async function from a cmd2 app."""

import asyncio
import concurrent.futures
import threading

import cmd2

_event_loop = None
_event_lock = threading.Lock()


def run_async(coro) -> concurrent.futures.Future:
    """
    Await a coroutine from a synchronous function/method.
    """

    global _event_loop  # noqa: PLW0603

    if _event_loop is None:
        with _event_lock:
            if _event_loop is None:
                _event_loop = asyncio.new_event_loop()
                thread = threading.Thread(
                    target=_event_loop.run_forever,
                    name='Async Runner',
                    daemon=True,
                )
                thread.start()

    return asyncio.run_coroutine_threadsafe(coro, _event_loop)


async def async_wait(duration: float) -> float:
    """
    Example async function that is called from a synchronous cmd2 command
    """
    await asyncio.sleep(duration)
    return duration


class AsyncCallExample(cmd2.Cmd):
    """
    A simple cmd2 application.
    Demonstrates how to run an async function from a cmd2 command.
    """

    def do_async_wait(self, _: str) -> None:
        """
        Waits asynchronously.  Example cmd2 command that calls an async function.
        """

        waitable = run_async(async_wait(0.1))
        self.poutput('Begin waiting...')
        # Wait for coroutine to complete and get its return value:
        res = waitable.result()
        self.poutput(f'Done waiting: {res}')
        return

    def do_hello_world(self, _: str) -> None:
        """
        Prints a simple greeting.  Just a typical (synchronous) cmd2 command
        """
        self.poutput('Hello World')


async def main() -> int:
    """
    Having this async ensures presence of the top level event loop.
    """
    app = AsyncCallExample()
    app.set_window_title("Call to an Async Function Test")
    return app.cmdloop()


if __name__ == '__main__':
    import sys

    sys.exit(asyncio.run(main(), debug=True))



#!/usr/bin/env python
"""A simple script to validate that a git tag matches a SemVer pattern."""

import re
import subprocess

SEMVER_SIMPLE = re.compile(r'(\d+)\.(\d+)\.(\d+)((a|b|rc)(\d+))?')
SEMVER_PATTERN = re.compile(
    r"""
        ^                                           # Start of the string
        v?                                          # Optional 'v' prefix (common in Git tags)
        (?P<major>0|[1-9]\d*)\.                     # Major version
        (?P<minor>0|[1-9]\d*)\.                     # Minor version
        (?P<patch>0|[1-9]\d*)                       # Patch version
        (?:-(?P<prerelease>                          # Optional pre-release section
            (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*) #  Identifier
            (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
        ))?
        (?:\+(?P<build>                              # Optional build metadata section
            [0-9a-zA-Z-]+                           #  Identifier
            (?:\.[0-9a-zA-Z-]+)*
        ))?
        $                                           # End of the string
    """,
    re.VERBOSE,
)


def get_current_tag() -> str:
    """Get current git tag."""
    try:
        # Gets the name of the latest tag reachable from the current commit
        result = subprocess.run(
            ['git', 'describe', '--exact-match', '--tags', '--abbrev=0'], capture_output=True, text=True, check=True
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError:
        print("Could not find a reachable tag.")
        return ''


def is_semantic_version(tag_name: str) -> bool:
    """Check if a given string complies with the semantic versioning 2.0.0 specification.

    Args:
        tag_name: The name of the Git tag to validate.

    Returns:
        bool: True if the tag is a valid semantic version, False otherwise.

    """
    # The regex pattern for semantic versioning 2.0.0 (source: https://semver.org/)
    semver_pattern = re.compile(
        r"""
        ^                                           # Start of the string
        v?                                          # Optional 'v' prefix (common in Git tags)
        (?P<major>0|[1-9]\d*)\.                     # Major version
        (?P<minor>0|[1-9]\d*)\.                     # Minor version
        (?P<patch>0|[1-9]\d*)                       # Patch version
        (?:-(?P<prerelease>                          # Optional pre-release section
            (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*) #  Identifier
            (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
        ))?
        (?:\+(?P<build>                              # Optional build metadata section
            [0-9a-zA-Z-]+                           #  Identifier
            (?:\.[0-9a-zA-Z-]+)*
        ))?
        $                                           # End of the string
    """,
        re.VERBOSE,
    )

    return bool(semver_pattern.match(tag_name))


if __name__ == '__main__':
    import sys

    git_tag = get_current_tag()
    if not git_tag:
        print('Git tag does not exist for current commit.')
        sys.exit(-1)

    if not is_semantic_version(git_tag):
        print(rf"Git tag '{git_tag}' is invalid according to SemVer.")
        sys.exit(-1)

    print(rf"Git tag '{git_tag}' is valid.")



"""Unit testing for cmd2/rich_utils.py module"""

from unittest import mock

import pytest
import rich.box
from rich.console import Console
from rich.style import Style
from rich.table import Table
from rich.text import Text

from cmd2 import (
    Cmd2Style,
    Color,
)
from cmd2 import rich_utils as ru

from .conftest import with_ansi_style


def test_cmd2_base_console() -> None:
    # Test the keyword arguments which are not allowed.
    with pytest.raises(TypeError) as excinfo:
        ru.Cmd2BaseConsole(color_system="auto")
    assert 'color_system' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        ru.Cmd2BaseConsole(force_terminal=True)
    assert 'force_terminal' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        ru.Cmd2BaseConsole(force_interactive=True)
    assert 'force_interactive' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        ru.Cmd2BaseConsole(theme=None)
    assert 'theme' in str(excinfo.value)


def test_indented_text() -> None:
    console = Console(width=20)

    # With an indention of 10, text will be evenly split across two lines.
    text = "A" * 20
    level = 10
    indented_text = ru.indent(text, level)

    with console.capture() as capture:
        console.print(indented_text)
    result = capture.get().splitlines()

    padding = " " * level
    expected_line = padding + ("A" * 10)
    assert result[0] == expected_line
    assert result[1] == expected_line


def test_indented_table() -> None:
    console = Console()

    level = 2
    table = Table("Column", box=rich.box.ASCII)
    table.add_row("Some Data")
    indented_table = ru.indent(table, level)

    with console.capture() as capture:
        console.print(indented_table)
    result = capture.get().splitlines()

    padding = " " * level
    assert result[0].startswith(padding + "+-----------+")
    assert result[1].startswith(padding + "| Column    |")
    assert result[2].startswith(padding + "|-----------|")
    assert result[3].startswith(padding + "| Some Data |")
    assert result[4].startswith(padding + "+-----------+")


@pytest.mark.parametrize(
    ('rich_text', 'string'),
    [
        (Text("Hello"), "Hello"),
        (Text("Hello\n"), "Hello\n"),
        # Test standard color support
        (Text("Standard", style="blue"), "\x1b[34mStandard\x1b[0m"),
        # Test 256-color support
        (Text("256-color", style=Color.NAVY_BLUE), "\x1b[38;5;17m256-color\x1b[0m"),
        # Test 24-bit color (TrueColor) support
        (Text("TrueColor", style="#123456"), "\x1b[38;2;18;52;86mTrueColor\x1b[0m"),
    ],
)
def test_rich_text_to_string(rich_text: Text, string: str) -> None:
    assert ru.rich_text_to_string(rich_text) == string


def test_set_theme() -> None:
    # Save a cmd2, rich-argparse, and rich-specific style.
    cmd2_style_key = Cmd2Style.ERROR
    argparse_style_key = "argparse.args"
    rich_style_key = "inspect.attr"

    orig_cmd2_style = ru.APP_THEME.styles[cmd2_style_key]
    orig_argparse_style = ru.APP_THEME.styles[argparse_style_key]
    orig_rich_style = ru.APP_THEME.styles[rich_style_key]

    # Overwrite these styles by setting a new theme.
    theme = {
        cmd2_style_key: Style(color=Color.CYAN),
        argparse_style_key: Style(color=Color.AQUAMARINE3, underline=True),
        rich_style_key: Style(color=Color.DARK_GOLDENROD, bold=True),
    }
    ru.set_theme(theme)

    # Verify theme styles have changed to our custom values.
    assert ru.APP_THEME.styles[cmd2_style_key] != orig_cmd2_style
    assert ru.APP_THEME.styles[cmd2_style_key] == theme[cmd2_style_key]

    assert ru.APP_THEME.styles[argparse_style_key] != orig_argparse_style
    assert ru.APP_THEME.styles[argparse_style_key] == theme[argparse_style_key]

    assert ru.APP_THEME.styles[rich_style_key] != orig_rich_style
    assert ru.APP_THEME.styles[rich_style_key] == theme[rich_style_key]


@with_ansi_style(ru.AllowStyle.ALWAYS)
def test_cmd2_base_console_print() -> None:
    """Test that Cmd2BaseConsole.print() correctly propagates formatting overrides to structured renderables."""
    from rich.rule import Rule

    # Create a console that defaults to no formatting
    console = ru.Cmd2BaseConsole(emoji=False, markup=False)

    # Use a Rule with emoji and markup in the title
    rule = Rule(title="[green]Success :1234:[/green]")

    with console.capture() as capture:
        # Override settings in the print() call
        console.print(rule, emoji=True, markup=True)

    result = capture.get()

    # Verify that the overrides were respected by checking for the emoji and the color code
    assert "🔢" in result
    assert "\x1b[32mSuccess" in result


@with_ansi_style(ru.AllowStyle.ALWAYS)
def test_cmd2_base_console_log() -> None:
    """Test that Cmd2BaseConsole.log() correctly propagates formatting overrides to structured renderables."""
    from rich.rule import Rule

    # Create a console that defaults to no formatting
    console = ru.Cmd2BaseConsole(emoji=False, markup=False)

    # Use a Rule with emoji and markup in the title
    rule = Rule(title="[green]Success :1234:[/green]")

    with console.capture() as capture:
        # Override settings in the log() call
        console.log(rule, emoji=True, markup=True)

    result = capture.get()

    # Verify that the formatting overrides were respected
    assert "🔢" in result
    assert "\x1b[32mSuccess" in result

    # Verify stack offset: the log line should point to this file, not rich_utils.py
    # Rich logs include the filename and line number on the right.
    assert "test_rich_utils.py" in result


@with_ansi_style(ru.AllowStyle.ALWAYS)
def test_cmd2_base_console_init_always_interactive_true() -> None:
    """Test Cmd2BaseConsole initialization when ALLOW_STYLE is ALWAYS and is_interactive is True."""
    with (
        mock.patch('rich.console.Console.__init__', return_value=None) as mock_base_init,
        mock.patch('cmd2.rich_utils.Console', autospec=True) as mock_detect_console_class,
    ):
        mock_detect_console = mock_detect_console_class.return_value
        mock_detect_console.is_interactive = True

        ru.Cmd2BaseConsole()

        # Verify arguments passed to super().__init__
        _, kwargs = mock_base_init.call_args
        assert kwargs['color_system'] == "truecolor"
        assert kwargs['force_terminal'] is True
        assert kwargs['force_interactive'] is True


@with_ansi_style(ru.AllowStyle.ALWAYS)
def test_cmd2_base_console_init_always_interactive_false() -> None:
    """Test Cmd2BaseConsole initialization when ALLOW_STYLE is ALWAYS and is_interactive is False."""
    with (
        mock.patch('rich.console.Console.__init__', return_value=None) as mock_base_init,
        mock.patch('cmd2.rich_utils.Console', autospec=True) as mock_detect_console_class,
    ):
        mock_detect_console = mock_detect_console_class.return_value
        mock_detect_console.is_interactive = False

        ru.Cmd2BaseConsole()

        _, kwargs = mock_base_init.call_args
        assert kwargs['color_system'] == "truecolor"
        assert kwargs['force_terminal'] is True
        assert kwargs['force_interactive'] is False


@with_ansi_style(ru.AllowStyle.TERMINAL)
def test_cmd2_base_console_init_terminal_true() -> None:
    """Test Cmd2BaseConsole initialization when ALLOW_STYLE is TERMINAL and it is a terminal."""
    with (
        mock.patch('rich.console.Console.__init__', return_value=None) as mock_base_init,
        mock.patch('cmd2.rich_utils.Console', autospec=True) as mock_detect_console_class,
    ):
        mock_detect_console = mock_detect_console_class.return_value
        mock_detect_console.is_terminal = True

        ru.Cmd2BaseConsole()

        _, kwargs = mock_base_init.call_args
        assert kwargs['color_system'] == "truecolor"
        assert kwargs['force_terminal'] is None
        assert kwargs['force_interactive'] is None


@with_ansi_style(ru.AllowStyle.TERMINAL)
def test_cmd2_base_console_init_terminal_false() -> None:
    """Test Cmd2BaseConsole initialization when ALLOW_STYLE is TERMINAL and it is not a terminal."""
    with (
        mock.patch('rich.console.Console.__init__', return_value=None) as mock_base_init,
        mock.patch('cmd2.rich_utils.Console', autospec=True) as mock_detect_console_class,
    ):
        mock_detect_console = mock_detect_console_class.return_value
        mock_detect_console.is_terminal = False

        ru.Cmd2BaseConsole()

        _, kwargs = mock_base_init.call_args
        assert kwargs['color_system'] is None
        assert kwargs['force_terminal'] is None
        assert kwargs['force_interactive'] is None


@with_ansi_style(ru.AllowStyle.NEVER)
def test_cmd2_base_console_init_never() -> None:
    """Test Cmd2BaseConsole initialization when ALLOW_STYLE is NEVER."""
    with mock.patch('rich.console.Console.__init__', return_value=None) as mock_base_init:
        ru.Cmd2BaseConsole()

        _, kwargs = mock_base_init.call_args
        assert kwargs['color_system'] is None
        assert kwargs['force_terminal'] is False
        assert kwargs['force_interactive'] is None



#!/usr/bin/env python3
"""A simple example cmd2 application demonstrating many common features.

Features demonstrated include all of the following:
 1) Colorizing/stylizing output
 2) Using multiline commands
 3) Persistent history
 4) How to run an initialization script at startup
 5) How to group and categorize commands when displaying them in help
 6) Opting-in to using the ipy command to run an IPython shell
 7) Allowing access to your application in py and ipy
 8) Displaying an intro banner upon starting your application
 9) Using a custom prompt
10) How to make custom attributes settable at runtime.
11) Shortcuts for commands
"""

import pathlib

from rich.style import Style

import cmd2
from cmd2 import (
    Color,
    stylize,
)


class BasicApp(cmd2.Cmd):
    """Cmd2 application to demonstrate many common features."""

    CUSTOM_CATEGORY = 'My Custom Commands'

    def __init__(self) -> None:
        """Initialize the cmd2 application."""
        # Startup script that defines a couple aliases for running shell commands
        alias_script = pathlib.Path(__file__).absolute().parent / '.cmd2rc'

        # Create a shortcut for one of our commands
        shortcuts = cmd2.DEFAULT_SHORTCUTS
        shortcuts.update({'&': 'intro'})
        super().__init__(
            include_ipy=True,
            multiline_commands=['echo'],
            persistent_history_file='cmd2_history.dat',
            shortcuts=shortcuts,
            startup_script=str(alias_script),
        )

        # Prints an intro banner once upon application startup
        self.intro = (
            stylize(
                'Welcome to cmd2!',
                style=Style(color=Color.GREEN1, bgcolor=Color.GRAY0, bold=True),
            )
            + ' Note the full Unicode support:  😇 💩'
        )

        # Show this as the prompt when asking for input
        self.prompt = 'myapp> '

        # Used as prompt for multiline commands after the first line
        self.continuation_prompt = '... '

        # Allow access to your application in py and ipy via self
        self.self_in_py = True

        # Set the default category name
        self.default_category = 'cmd2 Built-in Commands'

        # Color to output text in with echo command
        self.foreground_color = Color.CYAN.value

        # Make echo_fg settable at runtime
        fg_colors = [c.value for c in Color]
        self.add_settable(
            cmd2.Settable(
                'foreground_color',
                str,
                'Foreground color to use with echo command',
                self,
                choices=fg_colors,
            )
        )

    @cmd2.with_category(CUSTOM_CATEGORY)
    def do_intro(self, _: cmd2.Statement) -> None:
        """Display the intro banner."""
        self.poutput(self.intro)

    @cmd2.with_category(CUSTOM_CATEGORY)
    def do_echo(self, arg: cmd2.Statement) -> None:
        """Multiline command."""
        self.poutput(
            stylize(
                arg,
                style=Style(color=self.foreground_color),
            )
        )


if __name__ == '__main__':
    import sys

    app = BasicApp()
    sys.exit(app.cmdloop())



"""Test history functions of cmd2"""

import contextlib
import os
import tempfile
from unittest import (
    mock,
)

import pytest

import cmd2

from .conftest import (
    normalize,
    run_cmd,
)


def verify_hi_last_result(app: cmd2.Cmd, expected_length: int) -> None:
    """Verifies app.last_result when it contains a dictionary of HistoryItems"""
    assert len(app.last_result) == expected_length

    # Make sure last_result items match their respective indexes in the real history
    for key in app.last_result:
        # Keys in app.last_result are 1-based indexes into the app.history list
        assert app.last_result[key] == app.history[key - 1]


#
# readline tests
#
def test_readline_remove_history_item() -> None:
    from cmd2.rl_utils import (
        readline,
    )

    readline.clear_history()
    assert readline.get_current_history_length() == 0
    readline.add_history('this is a test')
    assert readline.get_current_history_length() == 1
    readline.remove_history_item(0)
    assert readline.get_current_history_length() == 0


#
# test History() class
#
@pytest.fixture
def hist():
    from cmd2.cmd2 import (
        History,
        HistoryItem,
    )
    from cmd2.parsing import (
        Statement,
    )

    return History(
        [
            HistoryItem(Statement('', raw='first')),
            HistoryItem(Statement('', raw='second')),
            HistoryItem(Statement('', raw='third')),
            HistoryItem(Statement('', raw='fourth')),
        ]
    )


# Represents the hist fixture's JSON
hist_json = (
    '{\n'
    '  "history_version": "1.0.0",\n'
    '  "history_items": [\n'
    '    {\n'
    '      "statement": {\n'
    '        "args": "",\n'
    '        "raw": "first",\n'
    '        "command": "",\n'
    '        "arg_list": [],\n'
    '        "multiline_command": "",\n'
    '        "terminator": "",\n'
    '        "suffix": "",\n'
    '        "pipe_to": "",\n'
    '        "output": "",\n'
    '        "output_to": ""\n'
    '      }\n'
    '    },\n'
    '    {\n'
    '      "statement": {\n'
    '        "args": "",\n'
    '        "raw": "second",\n'
    '        "command": "",\n'
    '        "arg_list": [],\n'
    '        "multiline_command": "",\n'
    '        "terminator": "",\n'
    '        "suffix": "",\n'
    '        "pipe_to": "",\n'
    '        "output": "",\n'
    '        "output_to": ""\n'
    '      }\n'
    '    },\n'
    '    {\n'
    '      "statement": {\n'
    '        "args": "",\n'
    '        "raw": "third",\n'
    '        "command": "",\n'
    '        "arg_list": [],\n'
    '        "multiline_command": "",\n'
    '        "terminator": "",\n'
    '        "suffix": "",\n'
    '        "pipe_to": "",\n'
    '        "output": "",\n'
    '        "output_to": ""\n'
    '      }\n'
    '    },\n'
    '    {\n'
    '      "statement": {\n'
    '        "args": "",\n'
    '        "raw": "fourth",\n'
    '        "command": "",\n'
    '        "arg_list": [],\n'
    '        "multiline_command": "",\n'
    '        "terminator": "",\n'
    '        "suffix": "",\n'
    '        "pipe_to": "",\n'
    '        "output": "",\n'
    '        "output_to": ""\n'
    '      }\n'
    '    }\n'
    '  ]\n'
    '}'
)


@pytest.fixture
def persisted_hist():
    from cmd2.cmd2 import (
        History,
        HistoryItem,
    )
    from cmd2.parsing import (
        Statement,
    )

    h = History(
        [
            HistoryItem(Statement('', raw='first')),
            HistoryItem(Statement('', raw='second')),
            HistoryItem(Statement('', raw='third')),
            HistoryItem(Statement('', raw='fourth')),
        ]
    )
    h.start_session()
    h.append(Statement('', raw='fifth'))
    h.append(Statement('', raw='sixth'))
    return h


def test_history_class_span(hist) -> None:
    span = hist.span('2..')
    assert len(span) == 3
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'

    span = hist.span('2:')
    assert len(span) == 3
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'

    span = hist.span('-2..')
    assert len(span) == 2
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'

    span = hist.span('-2:')
    assert len(span) == 2
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'

    span = hist.span('1..3')
    assert len(span) == 3
    assert span[1].statement.raw == 'first'
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'

    span = hist.span('1:3')
    assert len(span) == 3
    assert span[1].statement.raw == 'first'
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'

    span = hist.span('2:-1')
    assert len(span) == 3
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'

    span = hist.span('-3:4')
    assert len(span) == 3
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'

    span = hist.span('-4:-2')
    assert len(span) == 3
    assert span[1].statement.raw == 'first'
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'

    span = hist.span(':-2')
    assert len(span) == 3
    assert span[1].statement.raw == 'first'
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'

    span = hist.span('..-2')
    assert len(span) == 3
    assert span[1].statement.raw == 'first'
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'

    value_errors = ['fred', 'fred:joe', '2', '-2', 'a..b', '2 ..', '1 : 3', '1:0', '0:3']
    expected_err = "History indices must be positive or negative integers, and may not be zero."
    for tryit in value_errors:
        with pytest.raises(ValueError, match=expected_err):
            hist.span(tryit)


def test_persisted_history_span(persisted_hist) -> None:
    span = persisted_hist.span('2..')
    assert len(span) == 5
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'
    assert span[5].statement.raw == 'fifth'
    assert span[6].statement.raw == 'sixth'

    span = persisted_hist.span('-2..')
    assert len(span) == 2
    assert span[5].statement.raw == 'fifth'
    assert span[6].statement.raw == 'sixth'

    span = persisted_hist.span('1..3')
    assert len(span) == 3
    assert span[1].statement.raw == 'first'
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'

    span = persisted_hist.span('2:-1')
    assert len(span) == 5
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'
    assert span[5].statement.raw == 'fifth'
    assert span[6].statement.raw == 'sixth'

    span = persisted_hist.span('-3:4')
    assert len(span) == 1
    assert span[4].statement.raw == 'fourth'

    span = persisted_hist.span(':-2', include_persisted=True)
    assert len(span) == 5
    assert span[1].statement.raw == 'first'
    assert span[2].statement.raw == 'second'
    assert span[3].statement.raw == 'third'
    assert span[4].statement.raw == 'fourth'
    assert span[5].statement.raw == 'fifth'

    span = persisted_hist.span(':-2', include_persisted=False)
    assert len(span) == 1
    assert span[5].statement.raw == 'fifth'

    value_errors = ['fred', 'fred:joe', '2', '-2', 'a..b', '2 ..', '1 : 3', '1:0', '0:3']
    expected_err = "History indices must be positive or negative integers, and may not be zero."
    for tryit in value_errors:
        with pytest.raises(ValueError, match=expected_err):
            persisted_hist.span(tryit)


def test_history_class_get(hist) -> None:
    assert hist.get(1).statement.raw == 'first'
    assert hist.get(3).statement.raw == 'third'
    assert hist.get(-2) == hist[-2]
    assert hist.get(-1).statement.raw == 'fourth'

    with pytest.raises(IndexError):
        hist.get(0)

    with pytest.raises(IndexError):
        hist.get(5)


def test_history_str_search(hist) -> None:
    items = hist.str_search('ir')
    assert len(items) == 2
    assert items[1].statement.raw == 'first'
    assert items[3].statement.raw == 'third'

    items = hist.str_search('rth')
    assert len(items) == 1
    assert items[4].statement.raw == 'fourth'


def test_history_regex_search(hist) -> None:
    items = hist.regex_search('/i.*d/')
    assert len(items) == 1
    assert items[3].statement.raw == 'third'

    items = hist.regex_search('s[a-z]+ond')
    assert len(items) == 1
    assert items[2].statement.raw == 'second'


def test_history_max_length_zero(hist) -> None:
    hist.truncate(0)
    assert len(hist) == 0


def test_history_max_length_negative(hist) -> None:
    hist.truncate(-1)
    assert len(hist) == 0


def test_history_max_length(hist) -> None:
    hist.truncate(2)
    assert len(hist) == 2
    assert hist.get(1).statement.raw == 'third'
    assert hist.get(2).statement.raw == 'fourth'


def test_history_to_json(hist) -> None:
    assert hist_json == hist.to_json()


def test_history_from_json(hist) -> None:
    import json

    from cmd2.history import (
        History,
    )

    assert hist.from_json(hist_json) == hist

    # Test invalid JSON
    with pytest.raises(json.JSONDecodeError):
        hist.from_json("")

    # Send JSON with missing required element
    with pytest.raises(KeyError):
        hist.from_json("{}")

    # Create JSON with invalid history version
    backed_up_ver = History._history_version
    History._history_version = "BAD_VERSION"
    invalid_ver_json = hist.to_json()
    History._history_version = backed_up_ver

    expected_err = "Unsupported history file version: BAD_VERSION. This application uses version 1.0.0."
    with pytest.raises(ValueError, match=expected_err):
        hist.from_json(invalid_ver_json)


#
# test HistoryItem()
#
@pytest.fixture
def histitem():
    from cmd2.history import (
        HistoryItem,
    )
    from cmd2.parsing import (
        Statement,
    )

    statement = Statement(
        'history',
        raw='help history',
        command='help',
        arg_list=['history'],
    )
    return HistoryItem(statement)


@pytest.fixture
def parser():
    from cmd2.parsing import (
        StatementParser,
    )

    return StatementParser(
        terminators=[';', '&'],
        multiline_commands=['multiline'],
        aliases={
            'helpalias': 'help',
            '42': 'theanswer',
            'l': '!ls -al',
            'anothermultiline': 'multiline',
            'fake': 'run_pyscript',
        },
        shortcuts={'?': 'help', '!': 'shell'},
    )


def test_multiline_histitem(parser) -> None:
    from cmd2.history import (
        History,
    )

    line = 'multiline foo\nbar\n\n'
    statement = parser.parse(line)
    history = History()
    history.append(statement)
    assert len(history) == 1
    hist_item = history[0]
    assert hist_item.raw == line
    pr_lines = hist_item.pr(1).splitlines()
    assert pr_lines[0].endswith('multiline foo bar')


def test_multiline_with_quotes_histitem(parser) -> None:
    # Test that spaces and newlines in quotes are preserved
    from cmd2.history import (
        History,
    )

    line = 'Look, "There are newlines\n  and spaces  \n "\n in\nquotes.\n;\n'
    statement = parser.parse(line)
    history = History()
    history.append(statement)
    assert len(history) == 1
    hist_item = history[0]
    assert hist_item.raw == line

    # Since spaces and newlines in quotes are preserved, this history entry spans multiple lines.
    pr_lines = hist_item.pr(1).splitlines()
    assert pr_lines[0].endswith('Look, "There are newlines')
    assert pr_lines[1] == '  and spaces  '
    assert pr_lines[2] == ' " in quotes.;'


def test_multiline_histitem_verbose(parser) -> None:
    from cmd2.history import (
        History,
    )

    line = 'multiline foo\nbar\n\n'
    statement = parser.parse(line)
    history = History()
    history.append(statement)
    assert len(history) == 1
    hist_item = history[0]
    assert hist_item.raw == line
    pr_lines = hist_item.pr(1, verbose=True).splitlines()
    assert pr_lines[0].endswith('multiline foo')
    assert pr_lines[1] == 'bar'


def test_single_line_format_blank(parser) -> None:
    from cmd2.history import (
        single_line_format,
    )

    line = ""
    statement = parser.parse(line)
    assert single_line_format(statement) == line


def test_history_item_instantiate() -> None:
    from cmd2.history import (
        HistoryItem,
    )
    from cmd2.parsing import (
        Statement,
    )

    Statement(
        'history',
        raw='help history',
        command='help',
        arg_list=['history'],
    )
    with pytest.raises(TypeError):
        _ = HistoryItem()


def test_history_item_properties(histitem) -> None:
    assert histitem.raw == 'help history'
    assert histitem.expanded == 'help history'
    assert str(histitem) == 'help history'


#
# test history command
#
def test_base_history(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'shortcuts')
    out, _err = run_cmd(base_app, 'history')
    expected = normalize(
        """
    1  help
    2  shortcuts
"""
    )
    assert out == expected

    out, _err = run_cmd(base_app, 'history he')
    expected = normalize(
        """
    1  help
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 1)

    out, _err = run_cmd(base_app, 'history sh')
    expected = normalize(
        """
    2  shortcuts
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 1)


def test_history_script_format(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'shortcuts')
    out, _err = run_cmd(base_app, 'history -s')
    expected = normalize(
        """
help
shortcuts
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_with_string_argument(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'shortcuts')
    run_cmd(base_app, 'help history')
    out, _err = run_cmd(base_app, 'history help')
    expected = normalize(
        """
    1  help
    3  help history
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_expanded_with_string_argument(base_app) -> None:
    run_cmd(base_app, 'alias create sc shortcuts')
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'help history')
    run_cmd(base_app, 'sc')
    out, _err = run_cmd(base_app, 'history -v shortcuts')
    expected = normalize(
        """
    1  alias create sc shortcuts
    4  sc
    4x shortcuts
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_expanded_with_regex_argument(base_app) -> None:
    run_cmd(base_app, 'alias create sc shortcuts')
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'help history')
    run_cmd(base_app, 'sc')
    out, _err = run_cmd(base_app, 'history -v /sh.*cuts/')
    expected = normalize(
        """
    1  alias create sc shortcuts
    4  sc
    4x shortcuts
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_with_integer_argument(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'shortcuts')
    out, _err = run_cmd(base_app, 'history 1')
    expected = normalize(
        """
    1  help
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 1)


def test_history_with_integer_span(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'shortcuts')
    run_cmd(base_app, 'help history')
    out, _err = run_cmd(base_app, 'history 1..2')
    expected = normalize(
        """
    1  help
    2  shortcuts
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_with_span_start(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'shortcuts')
    run_cmd(base_app, 'help history')
    out, _err = run_cmd(base_app, 'history 2:')
    expected = normalize(
        """
    2  shortcuts
    3  help history
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_with_span_end(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'shortcuts')
    run_cmd(base_app, 'help history')
    out, _err = run_cmd(base_app, 'history :2')
    expected = normalize(
        """
    1  help
    2  shortcuts
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_with_span_index_error(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'help history')
    run_cmd(base_app, '!ls -hal :')
    expected_err = "History indices must be positive or negative integers, and may not be zero."
    with pytest.raises(ValueError, match=expected_err):
        base_app.onecmd('history "hal :"')


def test_history_output_file() -> None:
    app = cmd2.Cmd(multiline_commands=['alias'])
    run_cmd(app, 'help')
    run_cmd(app, 'shortcuts')
    run_cmd(app, 'help history')
    run_cmd(app, 'alias create my_alias history;')

    fd, fname = tempfile.mkstemp(prefix='', suffix='.txt')
    os.close(fd)
    run_cmd(app, f'history -o "{fname}"')
    assert app.last_result is True

    expected = normalize('help\nshortcuts\nhelp history\nalias create my_alias history;')
    with open(fname) as f:
        content = normalize(f.read())
    assert content == expected


def test_history_bad_output_file(base_app) -> None:
    run_cmd(base_app, 'help')
    run_cmd(base_app, 'shortcuts')
    run_cmd(base_app, 'help history')

    fname = os.path.join(os.path.sep, "fake", "fake", "fake")
    out, err = run_cmd(base_app, f'history -o "{fname}"')

    assert not out
    assert "Error saving" in err[0]
    assert base_app.last_result is False


def test_history_edit(monkeypatch) -> None:
    app = cmd2.Cmd(multiline_commands=['alias'])

    # Set a fake editor just to make sure we have one.  We aren't really
    # going to call it due to the mock
    app.editor = 'fooedit'

    # Mock out the run_editor call so we don't actually open an editor
    edit_mock = mock.MagicMock(name='run_editor')
    monkeypatch.setattr("cmd2.Cmd.run_editor", edit_mock)

    # Mock out the run_script call since the mocked edit won't produce a file
    run_script_mock = mock.MagicMock(name='do_run_script')
    monkeypatch.setattr("cmd2.Cmd.do_run_script", run_script_mock)

    # Put commands in history
    run_cmd(app, 'help')
    run_cmd(app, 'alias create my_alias history;')

    run_cmd(app, 'history -e 1:2')

    # Make sure both functions were called
    edit_mock.assert_called_once()
    run_script_mock.assert_called_once()


def test_history_run_all_commands(base_app) -> None:
    # make sure we refuse to run all commands as a default
    run_cmd(base_app, 'shortcuts')
    out, err = run_cmd(base_app, 'history -r')

    assert not out
    assert err[0].startswith("Cowardly refusing to run all")
    assert base_app.last_result is False


def test_history_run_one_command(base_app) -> None:
    out1, _err1 = run_cmd(base_app, 'help')
    out2, _err2 = run_cmd(base_app, 'history -r 1')
    assert out1 == out2
    assert base_app.last_result is True


def test_history_clear(mocker, hist_file) -> None:
    # Add commands to history
    app = cmd2.Cmd(persistent_history_file=hist_file)
    run_cmd(app, 'help')
    run_cmd(app, 'alias')

    # Make sure history has items
    out, err = run_cmd(app, 'history')
    assert out
    verify_hi_last_result(app, 2)

    # Clear the history
    run_cmd(app, 'history --clear')
    assert app.last_result is True

    # Make sure history is empty and its file is gone
    out, err = run_cmd(app, 'history')
    assert out == []
    assert not os.path.exists(hist_file)
    verify_hi_last_result(app, 0)

    # Clear the history again and make sure the FileNotFoundError from trying to delete missing history file is silent
    run_cmd(app, 'history --clear')
    assert app.last_result is True

    # Cause os.remove to fail and make sure error gets printed
    mock_remove = mocker.patch('os.remove')
    mock_remove.side_effect = OSError

    out, err = run_cmd(app, 'history --clear')
    assert out == []
    assert 'Error removing history file' in err[0]
    assert app.last_result is False


def test_history_verbose_with_other_options(base_app) -> None:
    # make sure -v shows a usage error if any other options are present
    options_to_test = ['-r', '-e', '-o file', '-t file', '-c', '-x']
    for opt in options_to_test:
        out, _err = run_cmd(base_app, 'history -v ' + opt)
        assert '-v cannot be used with any other options' in out
        assert base_app.last_result is False


def test_history_verbose(base_app) -> None:
    # validate function of -v option
    run_cmd(base_app, 'alias create s shortcuts')
    run_cmd(base_app, 's')
    out, _err = run_cmd(base_app, 'history -v')

    expected = normalize(
        """
    1  alias create s shortcuts
    2  s
    2x shortcuts
"""
    )
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_script_with_invalid_options(base_app) -> None:
    # make sure -s shows a usage error if -c, -r, -e, -o, or -t are present
    options_to_test = ['-r', '-e', '-o file', '-t file', '-c']
    for opt in options_to_test:
        out, _err = run_cmd(base_app, 'history -s ' + opt)
        assert '-s and -x cannot be used with -c, -r, -e, -o, or -t' in out
        assert base_app.last_result is False


def test_history_script(base_app) -> None:
    cmds = ['alias create s shortcuts', 's']
    for cmd in cmds:
        run_cmd(base_app, cmd)
    out, _err = run_cmd(base_app, 'history -s')
    assert out == cmds
    verify_hi_last_result(base_app, 2)


def test_history_expanded_with_invalid_options(base_app) -> None:
    # make sure -x shows a usage error if -c, -r, -e, -o, or -t are present
    options_to_test = ['-r', '-e', '-o file', '-t file', '-c']
    for opt in options_to_test:
        out, _err = run_cmd(base_app, 'history -x ' + opt)
        assert '-s and -x cannot be used with -c, -r, -e, -o, or -t' in out
        assert base_app.last_result is False


def test_history_expanded(base_app) -> None:
    # validate function of -x option
    cmds = ['alias create s shortcuts', 's']
    for cmd in cmds:
        run_cmd(base_app, cmd)
    out, _err = run_cmd(base_app, 'history -x')
    expected = ['    1  alias create s shortcuts', '    2  shortcuts']
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_history_script_expanded(base_app) -> None:
    # validate function of -s -x options together
    cmds = ['alias create s shortcuts', 's']
    for cmd in cmds:
        run_cmd(base_app, cmd)
    out, _err = run_cmd(base_app, 'history -sx')
    expected = ['alias create s shortcuts', 'shortcuts']
    assert out == expected
    verify_hi_last_result(base_app, 2)


def test_exclude_from_history(base_app) -> None:
    # Run history command
    run_cmd(base_app, 'history')
    verify_hi_last_result(base_app, 0)

    # Verify that the history is empty
    out, _err = run_cmd(base_app, 'history')
    assert out == []
    verify_hi_last_result(base_app, 0)

    # Now run a command which isn't excluded from the history
    run_cmd(base_app, 'help')

    # And verify we have a history now ...
    out, _err = run_cmd(base_app, 'history')
    expected = normalize("""    1  help""")
    assert out == expected
    verify_hi_last_result(base_app, 1)


#
# test history initialization
#
@pytest.fixture(scope="session")
def hist_file():
    fd, filename = tempfile.mkstemp(prefix='hist_file', suffix='.dat')
    os.close(fd)
    yield filename
    # teardown code
    with contextlib.suppress(FileNotFoundError):
        os.remove(filename)


def test_history_file_is_directory(capsys) -> None:
    with tempfile.TemporaryDirectory() as test_dir:
        # Create a new cmd2 app
        cmd2.Cmd(persistent_history_file=test_dir)
        _, err = capsys.readouterr()
        assert 'is a directory' in err


def test_history_can_create_directory(mocker) -> None:
    # Mock out atexit.register so the persistent file doesn't written when this function
    # exists because we will be deleting the directory it needs to go to.
    mocker.patch('atexit.register')

    # Create a temp path for us to use and let it get deleted
    with tempfile.TemporaryDirectory() as test_dir:
        pass
    assert not os.path.isdir(test_dir)

    # Add some subdirectories for the complete history file directory
    hist_file_dir = os.path.join(test_dir, 'subdir1', 'subdir2')
    hist_file = os.path.join(hist_file_dir, 'hist_file')

    # Make sure cmd2 creates the history file directory
    cmd2.Cmd(persistent_history_file=hist_file)
    assert os.path.isdir(hist_file_dir)

    # Cleanup
    os.rmdir(hist_file_dir)


def test_history_cannot_create_directory(mocker, capsys) -> None:
    mock_open = mocker.patch('os.makedirs')
    mock_open.side_effect = OSError

    hist_file_path = os.path.join('fake_dir', 'file')
    cmd2.Cmd(persistent_history_file=hist_file_path)
    _, err = capsys.readouterr()
    assert 'Error creating persistent history file directory' in err


def test_history_file_permission_error(mocker, capsys) -> None:
    mock_open = mocker.patch('builtins.open')
    mock_open.side_effect = PermissionError

    cmd2.Cmd(persistent_history_file='/tmp/doesntmatter')
    out, err = capsys.readouterr()
    assert not out
    assert 'Cannot read persistent history file' in err


def test_history_file_bad_compression(mocker, capsys) -> None:
    history_file = '/tmp/doesntmatter'
    with open(history_file, "wb") as f:
        f.write(b"THIS IS NOT COMPRESSED DATA")

    cmd2.Cmd(persistent_history_file=history_file)
    out, err = capsys.readouterr()
    assert not out
    assert 'Error decompressing persistent history data' in err


def test_history_file_bad_json(mocker, capsys) -> None:
    import lzma

    data = b"THIS IS NOT JSON"
    compressed_data = lzma.compress(data)

    history_file = '/tmp/doesntmatter'
    with open(history_file, "wb") as f:
        f.write(compressed_data)

    cmd2.Cmd(persistent_history_file=history_file)
    out, err = capsys.readouterr()
    assert not out
    assert 'Error processing persistent history data' in err


def test_history_populates_readline(hist_file) -> None:
    # - create a cmd2 with persistent history
    app = cmd2.Cmd(persistent_history_file=hist_file)
    run_cmd(app, 'help')
    run_cmd(app, 'shortcuts')
    run_cmd(app, 'shortcuts')
    run_cmd(app, 'alias')
    # call the private method which is registered to write history at exit
    app._persist_history()

    # see if history came back
    app = cmd2.Cmd(persistent_history_file=hist_file)
    assert len(app.history) == 4
    assert app.history.get(1).statement.raw == 'help'
    assert app.history.get(2).statement.raw == 'shortcuts'
    assert app.history.get(3).statement.raw == 'shortcuts'
    assert app.history.get(4).statement.raw == 'alias'

    # readline only adds a single entry for multiple sequential identical commands
    # so we check to make sure that cmd2 populated the readline history
    # using the same rules
    from cmd2.rl_utils import (
        readline,
    )

    assert readline.get_current_history_length() == 3
    assert readline.get_history_item(1) == 'help'
    assert readline.get_history_item(2) == 'shortcuts'
    assert readline.get_history_item(3) == 'alias'


#
# test cmd2's ability to write out history on exit
# we are testing the _persist_history() method, and
# we assume that the atexit module will call this method
# properly
#
def test_persist_history_ensure_no_error_if_no_histfile(base_app, capsys) -> None:
    # make sure if there is no persistent history file and someone
    # calls the private method call that we don't get an error
    base_app._persist_history()
    out, err = capsys.readouterr()
    assert not out
    assert not err


def test_persist_history_permission_error(hist_file, mocker, capsys) -> None:
    app = cmd2.Cmd(persistent_history_file=hist_file)
    run_cmd(app, 'help')
    mock_open = mocker.patch('builtins.open')
    mock_open.side_effect = PermissionError
    app._persist_history()
    out, err = capsys.readouterr()
    assert not out
    assert 'Cannot write persistent history file' in err



"""
The standard parser used by cmd2 built-in commands is Cmd2ArgumentParser.
The following code shows how to override it with your own parser class.
"""

import sys
from typing import NoReturn

from cmd2 import (
    Cmd2ArgumentParser,
    cmd2,
    set_default_argument_parser_type,
    styles,
    stylize,
)


# Since built-in commands rely on customizations made in Cmd2ArgumentParser,
# your custom parser class should inherit from Cmd2ArgumentParser.
class CustomParser(Cmd2ArgumentParser):
    """Overrides error method."""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)

    def error(self, message: str) -> NoReturn:
        """Custom override that applies custom formatting to the error message."""
        lines = message.split('\n')
        formatted_message = ''
        for linum, line in enumerate(lines):
            if linum == 0:
                formatted_message = 'Error: ' + line
            else:
                formatted_message += '\n       ' + line

        self.print_usage(sys.stderr)

        # Format errors with warning style
        formatted_message = stylize(
            formatted_message,
            style=styles.WARNING,
        )
        self.exit(2, f'{formatted_message}\n\n')


if __name__ == '__main__':
    import sys

    # Set the default parser type before instantiating app.
    set_default_argument_parser_type(CustomParser)

    app = cmd2.Cmd(include_ipy=True, persistent_history_file='cmd2_history.dat')
    app.self_in_py = True  # Enable access to "self" within the py command
    app.debug = True  # Show traceback if/when an exception occurs
    sys.exit(app.cmdloop())
